# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""AI Edge Quantizer API."""

from collections.abc import Iterable
import dataclasses
import json
import logging
import pathlib
from typing import Any, Optional, Union

import os
import io
from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import calibrator
from ai_edge_quantizer import default_policy
from ai_edge_quantizer import model_modifier
from ai_edge_quantizer import model_validator
from ai_edge_quantizer import params_generator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_quantizer.utils import validation_utils

Path = str | pathlib.Path
# Expose algorithm names to users.
AlgorithmName = algorithm_manager.AlgorithmName

_QuantRecipe = recipe_manager.ModelQuantizationRecipe
_TFLOpName = qtyping.TFLOperationName
_OpQuantizationConfig = qtyping.OpQuantizationConfig
_TensorQuantizationConfig = qtyping.TensorQuantizationConfig
_TensorTransformationParams = dict[str, qtyping.TensorTransformationParams]
_SignatureInput = dict[str, Any]  # input_argument_name -> tensor_value.
_CalibrationResult = dict[str, qtyping.QSV]


@dataclasses.dataclass(frozen=True)
class QuantizationResult:
  """Quantization result.

  Attributes:
    recipe: Quantization recipe.
    quantized_model: Quantized model.
  """

  recipe: _QuantRecipe
  quantized_model: Optional[bytearray]

  def save(
      self, save_folder: Path, model_name: str, overwrite: bool = False
  ) -> None:
    """Saves the quantized model and the quantization recipe.

    Args:
      save_folder: Path to the folder to save the quantized model and the
        quantization recipe.
      model_name: Name of the model.
      overwrite: Whether to overwrite the model if it already exists.

    Raises:
      RuntimeError: If no quantized model is available.
    """
    if not os.path.exists(save_folder):
      os.makedirs(save_folder)

    model_save_path = str(
        pathlib.Path(save_folder) / f'{model_name}.tflite'
    )
    self.export_model(model_save_path, overwrite)

    recipe_save_path = str(
        pathlib.Path(save_folder) / (model_name + '_recipe.json')
    )
    recipe = json.dumps(self.recipe)
    with open(recipe_save_path, 'w') as output_file_handle:
      output_file_handle.write(recipe)

  def export_model(self, filepath: Path, overwrite: bool = False) -> None:
    """Exports the quantized model to a .tflite flatbuffer.

    Args:
      filepath: Path (including file name) that the exported model should be
        serialized to.
      overwrite: Whether to overwrite the model if it already exists.

    Raises:
      RuntimeError: If no quantized model is available.
      ValueError: If the model already exists in the folder and overwrite is
        False.
    """
    if self.quantized_model is None:
      raise RuntimeError(
          'No quantized model to save. Make sure .quantize() is called.'
      )
    if os.path.exists(filepath):
      if overwrite:
        logging.warning(
            'The model %s already exists in the folder. Overwriting the model'
            ' since overwrite=True.',
            filepath,
        )
      else:
        raise ValueError(
            f'The model {filepath} already exists in the folder. Please'
            ' consider change the model name or specify overwrite=True to'
            ' overwrite the model if needed.'
        )
    with open(filepath, 'wb') as output_file_handle:
      output_file_handle.write(self.quantized_model)


class Quantizer:
  """AI Edge Quantizer API.

  Attributes:
    float_model_buffer: TFLite model bytearray.
    float_model: The `tf_flatbuffer_utils.ModelT` extracted from the
      `float_model_buffer`.
    quantization_recipe: Quantization recipe .json filepath or in loaded json
      format.
    previous_quantized_model_buffer: Optional previously quantized TFLite model
      bytearray. This is useful for validating a quantized model without
      quantizing it again.
  """

  def __init__(
      self,
      float_model: Union[Path, bytes, bytearray, memoryview],
      quantization_recipe: Optional[Union[Path, _QuantRecipe]] = None,
      previous_quantized_model: Optional[Union[Path, bytearray]] = None,
  ):
    """Initializes the quantizer.

    Args:
      float_model: Path to the float tflite model or model content in bytearray.
      quantization_recipe: Quantization recipe in .json filepath or loaded json
        format.
      previous_quantized_model: Path to an optional previously quantized tflite
        model. This is useful for validating a quantized model without
        quantizing it again.
    """
    # Load the `float_model` as a buffer.
    self._float_model_buffer = memoryview(
        tfl_flatbuffer_utils.get_model_content(float_model)
        if isinstance(float_model, (str, pathlib.Path))
        else float_model
    )
    if previous_quantized_model is not None:
      self.previous_quantized_model_buffer = memoryview(
          tfl_flatbuffer_utils.get_model_content(previous_quantized_model)
          if isinstance(previous_quantized_model, (str, pathlib.Path))
          else previous_quantized_model
      )
    else:
      self.previous_quantized_model_buffer = None

    # Extract the `float_model` from the buffer. Note that this will not
    # duplicate the model's data, i.e. all arrays are views on the data of the
    # underlying buffer.
    self._float_model: tfl_flatbuffer_utils.ModelT = (
        tfl_flatbuffer_utils.read_model(self._float_model_buffer)
    )

    self._recipe_manager: recipe_manager.RecipeManager = (
        recipe_manager.RecipeManager()
    )
    if quantization_recipe is not None:
      self.load_quantization_recipe(quantization_recipe)
    self._result: QuantizationResult = QuantizationResult([{}], None)
    self._quantize_called = False

  def load_quantization_recipe(self, recipe: Union[Path, _QuantRecipe]) -> None:
    """Loads a quantization recipe.

    The existing recipe will be overwritten.

    Args:
      recipe: Quantization recipe in json format.
    """
    if isinstance(recipe, (str, pathlib.Path)):
      with open(recipe) as json_file:
        recipe = json.load(json_file)
    self._recipe_manager.load_quantization_recipe(recipe)

  def load_config_policy(self, filename: Path) -> None:
    """Loads a JSON policy.

    The existing policy will be overwritten.

    Args:
      filename: Config policy filename.
    """
    with open(filename, 'r') as f:
      content = f.read()
      policy = default_policy.update_default_config_policy(content)

    # Register the policy for MIN_MAX_UNIFORM_QUANT algorithm.
    algorithm_manager.register_config_check_policy_func(
        AlgorithmName.MIN_MAX_UNIFORM_QUANT, policy
    )

  def get_quantization_recipe(self) -> _QuantRecipe:
    """Gets the quantization recipe.

    Returns:
      A quantization recipe.
    """
    return self._recipe_manager.get_quantization_recipe()

  def update_quantization_recipe(
      self,
      regex: str,
      operation_name: _TFLOpName,
      op_config: Optional[_OpQuantizationConfig] = None,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a quantization configuration to the recipe.

    Conflict arises when we are trying to set an operation under a certain regex
    which is already existed in the config dictionary. Under such circumstance,
    the new config is used to replace the previous one.

    We also have special treatment for _TFLOperationKey.ALL. If the new config
    is on _TFLOperationKey.ALL and there are existing op configs inside the same
    scope, we clear the previous configs and use _TFLOperationKey.ALL.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation. * for all supported TFLite
        operation.
      op_config: Quantization configuration which will be used to update the
        default configuration. None or empty dict means the default
        configuration will be used.
      algorithm_key: Algorithm key to be applied.
    """
    self._recipe_manager.add_quantization_config(
        regex, operation_name, op_config, algorithm_key
    )

  def add_dynamic_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      num_bits: int,
      granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a dynamic quantization configuration to the recipe.

    During dynamic quantization, activations are not processed by AEQ and
    remain in float format. The runtime kernel is expected to quantize these
    activations on-the-fly, as indicated by compute_precision=Integer and
    explicit_dequantize=False.

    The model quality may suffer due to the on-the-fly quantization. If quality
    is a concern, consider using weight-only
    quantization.

    Args:
      regex: Regular expression for layer name (op's output tensor name)
        matching.
      operation_name: Target TFLite operation.
      num_bits: Number of bits for quantization.
      granularity: Granularity of quantization.
      algorithm_key: Algorithm key to be applied.
    """
    self._recipe_manager.add_dynamic_config(
        regex, operation_name, num_bits, granularity, algorithm_key
    )

  def add_weight_only_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      num_bits: int,
      granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a weight only quantization configuration to the recipe.

    In weight-only quantization, weights are quantized, but the actual operation
    (op) computation remains in float. The quantized weight is explicitly
    dequantized before being fed into the op. This is achieved by inserting a
    dequantize op between the quantized weight and the consuming op. To enable
    this, both compute_precision will be set to Float and explicit_dequantize to
    True.

    Weight-only quantization is useful for reducing model size but may
    not decrease latency due to float computation. However, quantized model
    generally has better quality than other quantization options (e.g., dynamic
    range quantization) due to no loss of precision on activations. If latency
    is a concern, consider using dynamic quantization.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation.
      num_bits: Number of bits for quantization.
      granularity: Granularity of quantization.
      algorithm_key: Algorithm key to be applied.
    """
    self._recipe_manager.add_weight_only_config(
        regex, operation_name, num_bits, granularity, algorithm_key
    )

  def add_static_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      activation_num_bits: int,
      weight_num_bits: int,
      weight_granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a static quantization configuration to the recipe.

    In static quantization, both weights and activations are quantized. This
    requires a calibration step to determine the quantization parameters (e.g.,
    min/max ranges) for activations. The quantized model uses integer arithmetic
    for computations, which can lead to significant latency reductions.

    However, calibration is needed to determine the quantization parameters for
    activations, which requires sample data and may lead to quality loss. If
    there is no hardware requirement for full integer quantization, consider
    using dynamic quantization for simplicity.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation.
      activation_num_bits: Number of bits for activation quantization.
      weight_num_bits: Number of bits for weight quantization.
      weight_granularity: Granularity of weight quantization.
      algorithm_key: Algorithm key to be applied.
    """
    self._recipe_manager.add_static_config(
        regex,
        operation_name,
        activation_num_bits,
        weight_num_bits,
        weight_granularity,
        algorithm_key,
    )

  @property
  def need_calibration(self) -> bool:
    """Checks if the current recipe needs calibration."""
    return self._recipe_manager.need_calibration()

  def calibrate(
      self,
      calibration_data: dict[str, Iterable[_SignatureInput]],
      previous_calibration_result: Optional[_CalibrationResult] = None,
      num_threads: int = 16,
  ) -> _CalibrationResult:
    """Calibrates the float model (required by static range quantization).

    Args:
      calibration_data: Calibration data for a model signature.
      previous_calibration_result: Previous calibration result to be loaded. The
        calibration process will be resumed from the previous result.
      num_threads: Number of threads to use for calibration.

    Returns:
      Calibration result ({tensor_name: tensor QSVs (e.g.,min/max)}).

    Raises:
      ValueError: If the calibration result is insufficient.
    """
    if not self.need_calibration:
      return {}

    calib = calibrator.Calibrator(
        self._float_model_buffer, num_threads=num_threads
    )
    if previous_calibration_result is not None:
      calib.load_model_qsvs(previous_calibration_result)
    calib.calibrate(calibration_data, self._recipe_manager)
    return calib.get_model_qsvs()

  def _ensure_model_qsv_sufficient(
      self, calibration_result: _CalibrationResult
  ):
    """Checks if the calibration result has sufficient QSV."""

    # Find all tensor names with empty entries.
    empty_qsvs = [key for key, value in calibration_result.items() if not value]

    # Go over every signature and check if empty entry tensor belongs to it.
    tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(
        self._float_model_buffer
    )
    for signature_key in tfl_interpreter.get_signature_list():
      subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
          tfl_interpreter, signature_key
      )

      for tensor_detail in tfl_interpreter.get_tensor_details(subgraph_idx):
        tensor_name = tensor_detail['name']
        if tensor_name in empty_qsvs:
          raise ValueError(
              f'Missing QSVs (min/max) for tensor {tensor_name} in Signature'
              f" '{signature_key}'. Please check if Signature"
              f' {signature_key} has been calibrated.'
          )

  def quantize(
      self, calibration_result: Optional[_CalibrationResult] = None
  ) -> QuantizationResult:
    """Quantizes the float model.

    Args:
      calibration_result: Calibration result to be used for quantization (if
        needed, check with self.need_calibration).

    Returns:
      Quantization result.

    Raises:
      RuntimeError: If quantization recipe is empty.
    """
    self._quantize_called = True
    if calibration_result is not None:
      self._ensure_model_qsv_sufficient(calibration_result)

    if not self.get_quantization_recipe():
      raise RuntimeError('Can not quantize without a quantization recipe.')
    quant_params = self._get_quantization_params(calibration_result)
    quantized_model = self._get_quantized_model(quant_params)
    self._result = QuantizationResult(
        self.get_quantization_recipe(), quantized_model
    )
    return self._result

  def validate(
      self,
      test_data: Optional[dict[str, Iterable[_SignatureInput]]] = None,
      error_metrics: str = 'mse',
      use_xnnpack: bool = True,
      num_threads: int = 16,
      validate_output_tensors_only: bool = False,
  ) -> model_validator.ComparisonResult:
    """Numerical validation of the quantized model for a model signature.

    Side by side numerical comparison will be performed on all tensors in the
    quantized model against ones from the float model. If no test data is
    provided, random normal distributed data will be used. This test is intended
    to be SANITY check for the quality of the quantized model. End to end task
    specific test should be performed as the golden standard of the quantized
    model quality. The comparison result will be saved in json format if
    json_save_path is provided.

    Args:
      test_data: A dictionary of signature key and its correspending test input
        data that will be used for validation. If set to None, random normal
        distributed data will be used for all signatures in the model.
      error_metrics: Error metrics to be used for comparison.
      use_xnnpack: Whether to use the xnnpack library for validation.
      num_threads: Number of threads to use for validation.
      validate_output_tensors_only: If True, only compare output tensors.
        Otherwise, compare all tensors.

    Returns:
      The comparison result.
    """
    if test_data is None:
      # Create test data for all signatures in the model.
      test_data = tfl_interpreter_utils.create_random_normal_input_data(
          self._float_model_buffer, num_samples=1
      )
    if self._quantize_called:
      quantized_model = self._result.quantized_model
    else:
      quantized_model = self.previous_quantized_model_buffer

    if quantized_model is None:
      raise ValueError('No quantized model available to validate.')
    return model_validator.compare_model(
        self._float_model_buffer,
        quantized_model,
        test_data,
        error_metrics,
        validation_utils.get_validation_func(error_metrics),
        use_xnnpack=use_xnnpack,
        num_threads=num_threads,
        validate_output_tensors_only=validate_output_tensors_only,
    )

  def _get_quantization_params(
      self, calibration_result: Optional[_CalibrationResult] = None
  ) -> _TensorTransformationParams:
    """Gets the quantization parameters.

    Args:
      calibration_result: Calibration result to be used for quantization (if
        needed, check with self.need_calibration).

    Returns:
      A dictionary containing the quantization parameters.
    """
    params_generator_instance = params_generator.ParamsGenerator(
        self._float_model
    )
    return params_generator_instance.generate_quantization_parameters(
        self._recipe_manager, calibration_result
    )

  def _get_quantized_model(
      self, quant_params: _TensorTransformationParams
  ) -> bytearray:
    """Gets the quantized model.

    Args:
      quant_params: A dictionary containing the quantization parameters.

    Returns:
      The quantized model.
    """
    model_modifier_instance = model_modifier.ModelModifier(self._float_model)
    return model_modifier_instance.modify_model(quant_params)



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Python manager for transformations to be applied to TFlite models."""

from collections.abc import Sequence
from typing import Optional

import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.transformations import dequant_insert
from ai_edge_quantizer.transformations import duplicate_buffer
from ai_edge_quantizer.transformations import duplicate_tensor
from ai_edge_quantizer.transformations import insert_decomposed_hadamard_rotation
from ai_edge_quantizer.transformations import insert_hadamard_rotation
from ai_edge_quantizer.transformations import quant_insert
from ai_edge_quantizer.transformations import quantize_tensor
from ai_edge_quantizer.transformations import transformation_utils
from ai_edge_litert import schema_py_generated  # pylint: disable=g-direct-tensorflow-import


class TransformationPerformer:
  """Wrapper class for transformations.

  all transformations supported by the AI Edge Quantizer should be registered in
  the Transformation Performer.

  transformation to be appied to a tensor in the Instruction Generator is
  specified
  as a key. Given this key, the transformation performer will return a function
  that will apply the corresponding transformation on the graph

  A transformation is defined as a Callable that takes the following parameters:
  tensor_id: A tensor id that represents the tensor to be applied
  operatorCodes: list of OperatorCodesT from the source TFlite ModelT
  buffers: list of BufferT from the source TFLite ModelT
  subgraph: the specific subgraph where the transformation should be applied
  producer: the op index for the producer of the tensor
  consumers: a list of op index representing consumers to apply the change on
  quant_param: the quantization parameters in qtyping.UniformQuantParams
  And returns a qtyping.TransformationInfo which contains the index where the
  ops are added and how many ops are added

  Additionally, op additions must be consecutive

  this class is expected to be created by the Model Modifier and nothing else

  Model modifier would pass in a dict of transformations to be applied, this
  class will apply the transformations in a pre-determined static order
  """

  def __init__(self):
    """Initializes the TransformationPerformer."""
    self._transformation_registration = {
        qtyping.QuantTransformation.ADD_DEQUANTIZE: (
            dequant_insert.insert_dequant
        ),
        qtyping.QuantTransformation.QUANTIZE_TENSOR: (
            quantize_tensor.quantize_tensor
        ),
        qtyping.QuantTransformation.EMULATED_SUBCHANNEL: (
            transformation_utils.raise_deprecated_error
        ),
        qtyping.QuantTransformation.ADD_QUANTIZE: quant_insert.insert_quant,
        qtyping.QuantTransformation.DUPLICATE_BUFFER: (
            duplicate_buffer.duplicate_buffer
        ),
        qtyping.QuantTransformation.DUPLICATE_TENSOR: (
            duplicate_tensor.duplicate_tensor
        ),
        qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION: (
            insert_hadamard_rotation.insert_hadamard_rotation
        ),
        qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION: (
            insert_decomposed_hadamard_rotation.insert_decomposed_hadamard_rotation
        ),
    }
    # transformations are seprated in two categories:
    # op_insertion_transformations are transformations that only insert ops
    # into the graph, whereas op_replacement_transformations will replace one op
    # with a pattern
    self._op_insertion_transformations = set([
        qtyping.QuantTransformation.ADD_DEQUANTIZE,
        qtyping.QuantTransformation.QUANTIZE_TENSOR,
        qtyping.QuantTransformation.ADD_QUANTIZE,
        qtyping.QuantTransformation.DUPLICATE_BUFFER,
        qtyping.QuantTransformation.DUPLICATE_TENSOR,
        qtyping.QuantTransformation.INSERT_HADAMARD_ROTATION,
        qtyping.QuantTransformation.INSERT_DECOMPOSED_HADAMARD_ROTATION,
    ])
    self._op_replacement_transformations = set(
        [qtyping.QuantTransformation.EMULATED_SUBCHANNEL]
    )
    self._original_op_id_map = []
    self._added_op_id_map = []

  def _create_op_id_map(self, tflite_model: schema_py_generated.ModelT):
    """init the original op_id to modified op_id map.

    At the beginning the graph has not been updated, so op_id maps to it's
    current id.

    Args:
      tflite_model: the model we're create op_id mapping

    Returns:
      None, modifies self._original_op_id_map inplace
    """
    for subgraph in tflite_model.subgraphs:
      self._original_op_id_map.append(list(range(len(subgraph.operators))))
      self._added_op_id_map.append([])

  def _update_op_id_map(
      self, subgraph_id: int, original_op_id: int, num_ops_added: int
  ):
    """Update the mapping between the original op id and modified op ids.

    Args:
      subgraph_id: the index of subgraph that we're interested in
      original_op_id: the original id for which the first op is added
      num_ops_added: the number of ops added starting from the op id

    Returns:
      None, modify self._original_op_id_map
    """
    np_op_id_map = np.array(self._original_op_id_map[subgraph_id])
    np_op_id_map[original_op_id:] += num_ops_added
    self._original_op_id_map[subgraph_id] = np_op_id_map.tolist()

  def _update_instructions(
      self,
      prev_transformation_index: int,
      transformations: list[qtyping.TransformationInst],
      subgraph_id: int,
      trans_info: qtyping.TransformationInfo,
  ) -> None:
    """Update the instructions in-place after the graph is modified.

    After an op is inserted or a tensor is duplicated, the topology is changed
    and this may impact the following transformation to be applied. So we need
    to update instructions that have yet to be applied.

    Args:
      prev_transformation_index: The index of the last applied transformation.
      transformations: The list of transformations we're applying.
      subgraph_id: The subgraph where the provided instructions belong to.
      trans_info: Transformation info returned by a transformation.
    """
    prev_transformation = transformations[prev_transformation_index]
    is_prev_not_duplicate_tensor = (
        prev_transformation.transformation
        != qtyping.QuantTransformation.DUPLICATE_TENSOR
    )
    was_op_added = trans_info.num_ops_added > 0
    if not was_op_added and is_prev_not_duplicate_tensor:
      return

    if was_op_added:
      self._added_op_id_map[subgraph_id].append(
          trans_info.op_id + trans_info.num_ops_added - 1
      )

    for transformations_index in range(
        prev_transformation_index + 1, len(transformations)
    ):
      transformation = transformations[transformations_index]
      for consumer_index in transformation.consumers:
        # If the consumer needs to use newly added ops, then the new added op
        # index needs to be outside of the range of the orignal op ids.
        if consumer_index in prev_transformation.consumers:
          if was_op_added:
            transformation.producer = (
                len(self._original_op_id_map[subgraph_id])
                + len(self._added_op_id_map[subgraph_id])
                - 1
            )
          transformation.tensor_id = trans_info.output_tensor_id

  def _get_updated_producer_id(
      self, original_producer_id: int, subgraph_id: int
  ) -> int:
    """Update the producer of a transformation instruction."""
    if original_producer_id is None or original_producer_id < 0:
      producer = -1
    elif original_producer_id < len(self._original_op_id_map[subgraph_id]):
      producer = self._original_op_id_map[subgraph_id][original_producer_id]
    else:
      # If the producer id is not in the original op map, it's an added op,
      # go the added op map to find the producer.
      producer = self._added_op_id_map[subgraph_id][
          original_producer_id - len(self._original_op_id_map[subgraph_id])
      ]
    return producer

  def _get_updated_consumer_ids(
      self,
      original_consumer_ids: list[int],
      subgraph_id: int,
  ) -> list[int]:
    """Update the consumers of a transformation instruction."""
    consumers = []
    for original_op_id in original_consumer_ids:
      new_consumer_id = (
          -1
          if original_op_id == -1
          else self._original_op_id_map[subgraph_id][original_op_id]
      )
      consumers.append(new_consumer_id)
    return consumers

  def _apply_single_transformation(
      self,
      transformation_inst: qtyping.TensorTransformationInsts,
      transformation_index: int,
      tflite_model: schema_py_generated.ModelT,
  ):
    """Apply a single transformation.

    Args:
      transformation_inst: a TensorTransformationInsts type that contains all
        transformations on a tensor
      transformation_index: the index of the transformation to be applied
      tflite_model: source tflite model to be updated

    Returns:
      None, update the transformation_inst & tflite_model in place
    """
    instruction = transformation_inst.instructions[transformation_index]
    producer = self._get_updated_producer_id(
        instruction.producer, transformation_inst.subgraph_id
    )
    consumers = self._get_updated_consumer_ids(
        instruction.consumers, transformation_inst.subgraph_id
    )
    trans_info = self._transformation_registration[instruction.transformation](
        transformation_utils.TransformationInput(
            instruction.tensor_id,
            tflite_model.operatorCodes,
            tflite_model.buffers,
            tflite_model.subgraphs[transformation_inst.subgraph_id],
            producer,
            consumers,
            instruction.parameters,
        )
    )
    self._update_instructions(
        transformation_index,
        transformation_inst.instructions,
        transformation_inst.subgraph_id,
        trans_info,
    )
    self._update_op_id_map(
        transformation_inst.subgraph_id,
        # The added op must be right before the most immediate consumer, unless
        # the consumer is the graph output (id=-1), then use the producer's
        # index instead.
        min(instruction.consumers)
        if min(instruction.consumers) >= 0
        else instruction.producer + 1,
        trans_info.num_ops_added,
    )

  def _apply_transformations(
      self,
      transformation_inst: qtyping.TensorTransformationInsts,
      tflite_model: schema_py_generated.ModelT,
  ):
    """Apply all transformations for a tensor.

    transformations are separated in two types and applied separately in two
    different passes

    Args:
      transformation_inst: a TensorTransformationInsts type that contains all
        transformation on a tensor
      tflite_model: source tflite model to be updated

    Returns:
      None, update the transformation_inst & tflite_model in place
    """
    # pass 1: apply all the op insertion transformation, because op replacement
    # may remove consumer or producer of some tensors
    for index, instruction in enumerate(transformation_inst.instructions):
      if instruction.transformation in self._op_insertion_transformations:
        self._apply_single_transformation(
            transformation_inst, index, tflite_model
        )
    # pass 2: apply all the op replacement transformation
    for index, instruction in enumerate(transformation_inst.instructions):
      if instruction.transformation in self._op_replacement_transformations:
        self._apply_single_transformation(
            transformation_inst, index, tflite_model
        )

  def transform_graph(
      self,
      transformation_instructions: dict[str, qtyping.TensorTransformationInsts],
      tflite_model: schema_py_generated.ModelT,
      tensor_processing_order: Optional[Sequence[str]] = None,
  ) -> None:
    """Apply all transformations to the given tflite_model in place.

    Args:
      transformation_instructions: Mapping from tensor name to its
        transformation instructions, produced by
        transformation_instruction_generator.
      tflite_model: The tflite model to apply quantization to.
      tensor_processing_order: The order of tensors to process. If not provided,
        the order will be inferred from `transformation_instructions`.
    """
    self._original_op_id_map = []
    self._added_op_id_map = []
    self._create_op_id_map(tflite_model)
    if tensor_processing_order is None:
      tensor_processing_order = transformation_instructions.keys()
    for tensor_name in tensor_processing_order:
      self._apply_transformations(
          transformation_instructions[tensor_name], tflite_model
      )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""flatbuffer utils for the Quantizer."""

import logging
import mmap
import os
import pathlib
from typing import Any, Optional, Union

import immutabledict
import numpy as np

import os
import io
from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_litert import schema_py_generated as schema  # pylint:disable=g-direct-tensorflow-import


# Export some common schema types.
ModelT = schema.ModelT


_TFLOpName = qtyping.TFLOperationName

Path = Union[str, pathlib.Path]

TFL_OP_NAME_TO_CODE = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: schema.BuiltinOperator.FULLY_CONNECTED,
    _TFLOpName.BATCH_MATMUL: schema.BuiltinOperator.BATCH_MATMUL,
    _TFLOpName.CONV_2D: schema.BuiltinOperator.CONV_2D,
    _TFLOpName.DEPTHWISE_CONV_2D: schema.BuiltinOperator.DEPTHWISE_CONV_2D,
    _TFLOpName.CONV_2D_TRANSPOSE: schema.BuiltinOperator.TRANSPOSE_CONV,
    _TFLOpName.EMBEDDING_LOOKUP: schema.BuiltinOperator.EMBEDDING_LOOKUP,
    _TFLOpName.SOFTMAX: schema.BuiltinOperator.SOFTMAX,
    _TFLOpName.AVERAGE_POOL_2D: schema.BuiltinOperator.AVERAGE_POOL_2D,
    _TFLOpName.RESHAPE: schema.BuiltinOperator.RESHAPE,
    _TFLOpName.TANH: schema.BuiltinOperator.TANH,
    _TFLOpName.TRANSPOSE: schema.BuiltinOperator.TRANSPOSE,
    _TFLOpName.GELU: schema.BuiltinOperator.GELU,
    _TFLOpName.ADD: schema.BuiltinOperator.ADD,
    _TFLOpName.SUB: schema.BuiltinOperator.SUB,
    _TFLOpName.MUL: schema.BuiltinOperator.MUL,
    _TFLOpName.MEAN: schema.BuiltinOperator.MEAN,
    _TFLOpName.RSQRT: schema.BuiltinOperator.RSQRT,
    _TFLOpName.CONCATENATION: schema.BuiltinOperator.CONCATENATION,
    _TFLOpName.STRIDED_SLICE: schema.BuiltinOperator.STRIDED_SLICE,
    _TFLOpName.SPLIT: schema.BuiltinOperator.SPLIT,
    _TFLOpName.LOGISTIC: schema.BuiltinOperator.LOGISTIC,
    _TFLOpName.SLICE: schema.BuiltinOperator.SLICE,
    _TFLOpName.SUM: schema.BuiltinOperator.SUM,
    _TFLOpName.SELECT: schema.BuiltinOperator.SELECT,
    _TFLOpName.SELECT_V2: schema.BuiltinOperator.SELECT_V2,
    _TFLOpName.STABLEHLO_COMPOSITE: schema.BuiltinOperator.STABLEHLO_COMPOSITE,
    _TFLOpName.DYNAMIC_UPDATE_SLICE: (
        schema.BuiltinOperator.DYNAMIC_UPDATE_SLICE
    ),
    _TFLOpName.PAD: schema.BuiltinOperator.PAD,
    _TFLOpName.SQUARED_DIFFERENCE: schema.BuiltinOperator.SQUARED_DIFFERENCE,
    _TFLOpName.MAX_POOL_2D: schema.BuiltinOperator.MAX_POOL_2D,
    _TFLOpName.RESIZE_BILINEAR: schema.BuiltinOperator.RESIZE_BILINEAR,
    _TFLOpName.RESIZE_NEAREST_NEIGHBOR: (
        schema.BuiltinOperator.RESIZE_NEAREST_NEIGHBOR
    ),
    _TFLOpName.GATHER_ND: schema.BuiltinOperator.GATHER_ND,
    _TFLOpName.PACK: schema.BuiltinOperator.PACK,
    _TFLOpName.UNPACK: schema.BuiltinOperator.UNPACK,
    _TFLOpName.DIV: schema.BuiltinOperator.DIV,
    _TFLOpName.BROADCAST_TO: schema.BuiltinOperator.BROADCAST_TO,
    _TFLOpName.SQRT: schema.BuiltinOperator.SQRT,
    _TFLOpName.GATHER: schema.BuiltinOperator.GATHER,
    _TFLOpName.HARD_SWISH: schema.BuiltinOperator.HARD_SWISH,
    _TFLOpName.MAXIMUM: schema.BuiltinOperator.MAXIMUM,
    _TFLOpName.PADV2: schema.BuiltinOperator.PADV2,
    _TFLOpName.REDUCE_MIN: schema.BuiltinOperator.REDUCE_MIN,
    _TFLOpName.EQUAL: schema.BuiltinOperator.EQUAL,
    _TFLOpName.NOT_EQUAL: schema.BuiltinOperator.NOT_EQUAL,
    _TFLOpName.MIRROR_PAD: schema.BuiltinOperator.MIRROR_PAD,
    _TFLOpName.SPACE_TO_DEPTH: schema.BuiltinOperator.SPACE_TO_DEPTH,
    _TFLOpName.RELU: schema.BuiltinOperator.RELU,
})

TFL_OP_CODE_TO_NAME = immutabledict.immutabledict(
    dict((reversed(item) for item in TFL_OP_NAME_TO_CODE.items()))
)

# Quantized dimension for per-channel quantization.
# See https://www.tensorflow.org/lite/performance/quantization_spec.
TFL_OP_TO_WEIGHT_QUANTIZED_DIM = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: 0,
    _TFLOpName.DEPTHWISE_CONV_2D: 3,
    _TFLOpName.CONV_2D: 0,
    _TFLOpName.EMBEDDING_LOOKUP: 0,
    _TFLOpName.CONV_2D_TRANSPOSE: 0,
})

TFL_OP_TO_BLOCKWISE_WEIGHT_QUANTIZED_DIM = immutabledict.immutabledict({
    _TFLOpName.FULLY_CONNECTED: 1,
    _TFLOpName.EMBEDDING_LOOKUP: 1,
})

NUM_TFL_DATATYPES = 18
TENSOR_CODE_TO_TYPE = {}
for dtype_code in range(NUM_TFL_DATATYPES):
  TENSOR_CODE_TO_TYPE[dtype_code] = flatbuffer_utils.type_to_name(dtype_code)
TENSOR_CODE_TO_TYPE = immutabledict.immutabledict(TENSOR_CODE_TO_TYPE)
TENSOR_TYPE_TO_CODE = immutabledict.immutabledict(  # pytype: disable=wrong-arg-types
    (reversed(item) for item in TENSOR_CODE_TO_TYPE.items())
)

# Expose functions in litert.python.tools.flatbuffer_utils
write_model = flatbuffer_utils.write_model


def read_model(
    tflite_model: Union[Path, bytearray, bytes, memoryview],
) -> schema.ModelT:
  """Read and convert the TFLite model into a flatbuffer object.

  Args:
    tflite_model: TFLite model path or bytearray.

  Raises:
    ValueError: Unsupported tflite_model type.

  Returns:
    flatbuffer_model: the flatbuffer_model.
  """
  if isinstance(tflite_model, Path):
    return flatbuffer_utils.read_model(tflite_model)
  elif isinstance(tflite_model, (bytes, bytearray, memoryview)):
    return flatbuffer_utils.read_model_from_bytearray(tflite_model)
  else:
    raise ValueError(
        "Unsupported tflite_model type: %s" % type(tflite_model).__name__
    )


def get_model_content(tflite_path: Path) -> memoryview:
  """Get the model content (read-only bytes) from the path.

  Args:
    tflite_path: Path to the .tflite.

  Returns:
    The model bytes.
  """
  model_bytes = None

  # Try to mmap the file first if it is local.
  if (fd := os.open(tflite_path, os.O_RDONLY)) >= 0:
    try:
      model_bytes = mmap.mmap(fd, 0, flags=mmap.MAP_SHARED, prot=mmap.PROT_READ)
    except IOError as e:
      print(f"Mapping model file {tflite_path} failed with exception: {e}.")
    os.close(fd)

  # If mapping failed, go at it conventionally.
  if model_bytes is None:
    with open(tflite_path, "rb") as tflite_file:
      model_bytes = tflite_file.read()

  return memoryview(model_bytes)


def get_model_buffer(tflite_path: Path) -> bytearray:
  """Get a mutable model buffer from the path.

  Args:
    tflite_path: path to the .tflite.

  Returns:
    model_buffer: the model buffer.
  """
  model_bytearray = None

  # Try to mmap the file first if it is local.
  try:
    if (fd := os.open(tflite_path, os.O_RDONLY)) >= 0:
      try:
        model_mmap = mmap.mmap(
            fd, 0, flags=mmap.MAP_SHARED, prot=mmap.PROT_READ
        )
        model_bytearray = bytearray(model_mmap[:])
      except IOError as e:
        print(f"Mapping model file {tflite_path} failed with exception: {e}.")
      os.close(fd)
  except RuntimeError:
    pass

  # If mapping failed, go at it conventionally.
  if model_bytearray is None:
    with open(tflite_path, "rb") as tflite_file:
      model_bytearray = bytearray(tflite_file.read())

  return model_bytearray


def parse_op_tensors(op: Any, subgraph_tensors: list[Any]) -> list[Any]:
  """Parse the op tensors.

  Args:
    op: the op that need to be parsed.
    subgraph_tensors: list of tensors in the subgraph.

  Returns:
    tensors: list of tensors that are associated with the op.
  """

  tensors = []
  for tensor_idx in list(op.outputs) + list(op.inputs):
    if tensor_idx != -1:
      tensors.append(subgraph_tensors[tensor_idx])
  return tensors


def parse_fc_bmm_conv_tensors(
    op: Any,
    subgraph_tensors: list[Any],
    input_index: int = 0,
    weight_index: int = 1,
    bias_index: int = 2,
    output_index: int = 0,
) -> tuple[Any, Any, Any, Any]:
  """Parse tensors in FullyConnected, BatchMatmul, and Convolutions.

  Args:
    op: the TFLite op, must be fully_connected, batch_matmul, or convolution.
    subgraph_tensors: tensors in the subgraph.
    input_index: index for the input tensor.
    weight_index: index for the weight tensor.
    bias_index: index for the bias tensor.
    output_index: index for the output tensor.

  Returns:
    input_tensor, weight_tensor, bias_tensor, output_tensor
  """

  input_tensor = subgraph_tensors[op.inputs[input_index]]
  weight_tensor = subgraph_tensors[op.inputs[weight_index]]
  bias_tensor = None
  if bias_index < len(op.inputs) and op.inputs[bias_index] != -1:
    bias_tensor = subgraph_tensors[op.inputs[bias_index]]
  output_tensor = subgraph_tensors[op.outputs[output_index]]
  return input_tensor, weight_tensor, bias_tensor, output_tensor


# flatbuffer_model has Any type since litert.python.tools.flatbuffer_utils
# is not type annotated.
def buffer_to_tensors(flatbuffer_model: Any) -> dict[int, list[Any]]:
  """Returns a map from buffer id to tensors that use it."""
  buffer_to_tensor_map = {}
  for subgraph in flatbuffer_model.subgraphs:
    for op in subgraph.operators:
      for tensor in parse_op_tensors(op, subgraph.tensors):
        if tensor.buffer not in buffer_to_tensor_map:
          buffer_to_tensor_map[tensor.buffer] = []
        if tensor not in buffer_to_tensor_map[tensor.buffer]:
          buffer_to_tensor_map[tensor.buffer].append(tensor)
  return buffer_to_tensor_map


def get_tensor_name(tensor: Any) -> str:
  """Get the tensor name for a fb tensor.

  Args:
    tensor: tensor in flatbuffer.

  Returns:
    tensor_name: name of the buffer
  """
  return tensor.name.decode("utf-8")


def get_tensor_data(tensor: Any, buffers: list[Any]) -> Optional[np.ndarray]:
  """Get the tensor data.

  Args:
    tensor: tensor in flatbuffer.
    buffers: list of buffers

  Returns:
    tensor_data: data inside the tensor
  """
  tensor_buffer = buffers[tensor.buffer]
  buffer_data = tensor_buffer.data
  if buffer_data is None:
    return None
  data = np.frombuffer(
      buffer_data, dtype=TENSOR_CODE_TO_TYPE[tensor.type].lower()
  )
  if tensor.shape is not None:
    data = np.reshape(data, tensor.shape)
  return data


def has_same_quantization(tensor1: Any, tensor2: Any) -> bool:
  """Check if two tensors have the same quantization.

  Args:
    tensor1: tensor in flatbuffer.
    tensor2: tensor in flatbuffer.

  Returns:
    True if two tensors have the same quantization.
  """

  def to_tuple(val):
    if val is None:
      val = []
    return tuple(val)

  same_type = tensor1.type == tensor2.type

  # Return True if both tensors are not quantized.
  if tensor1.quantization is None and tensor2.quantization is None:
    return True
  if tensor1.quantization.scale is None and tensor2.quantization.scale is None:
    return True

  same_scale = to_tuple(tensor1.quantization.scale) == to_tuple(
      tensor2.quantization.scale
  )
  same_zero_point = to_tuple(tensor1.quantization.zeroPoint) == to_tuple(
      tensor2.quantization.zeroPoint
  )
  same_quantized_dimension = (
      tensor1.quantization.quantizedDimension
      == tensor2.quantization.quantizedDimension
  )
  return (
      same_type and same_scale and same_zero_point and same_quantized_dimension
  )


def is_float_model(flatbuffer_model: Any) -> bool:
  """Checks that the model is float and not already quantized."""
  for subgraph in flatbuffer_model.subgraphs:
    for tensor in subgraph.tensors:
      if tensor.quantization is None:
        continue
      if tensor.quantization.scale is not None:
        return False
  return True


def get_subgraph_input_output_operators(
    subgraph: Any,
) -> list[qtyping.IOOperator]:
  """Get the input/output operators for the subgraph.

  Args:
    subgraph: The subgraph object.

  Returns:
    Input and output operators for the subgraph.
  """
  input_op = qtyping.IOOperator(
      inputs=[],
      outputs=subgraph.inputs,
      op_key=qtyping.TFLOperationName.INPUT,
  )
  output_op = qtyping.IOOperator(
      inputs=subgraph.outputs,
      outputs=[],
      op_key=qtyping.TFLOperationName.OUTPUT,
  )
  return [input_op, output_op]


def get_op_side_effect_subgraphs(
    op: Union[schema.Operator, schema.OperatorT],
) -> list[int]:
  """Get indices of any subgraphs invoked as a side effect of the operator.

  Args:
    op: The operator object.

  Returns:
    A list of subgraph indices invoked by the operator. Empty if the operator
    does not invoke any subgraphs.
  """
  if opts := flatbuffer_utils.get_options_as(
      op, schema.StableHLOCompositeOptionsT
  ):
    return [opts.decompositionSubgraphIndex]
  # Can add other nested ops here (control flow ops, etc).
  return []


def get_op_name_by_index(
    flatbuffer_model: Any, subgraph_id: int, op_index: int
) -> str:
  """Get the op name from the flatbuffer model."""
  op = flatbuffer_model.subgraphs[subgraph_id].operators[op_index]
  builtin_code = flatbuffer_model.operatorCodes[op.opcodeIndex].builtinCode
  return TFL_OP_CODE_TO_NAME[builtin_code]


def get_op_scope(
    op: Any, subgraph_tensors: list[Any], max_length: int = 10000
) -> str:
  """Get the op scope.

  Op scope is defined by the output tensor names (following ModelExplorer). If
  no output tensors are present, the input tensor names are used instead.

  Args:
    op: The op that needs to be parsed.
    subgraph_tensors: Tensors in the subgraph.
    max_length: The maximum length of the scope string. If the scope string is
      longer than this length, it will be truncated to this length to avoid
      overwhelming regex matching engine.

  Returns:
    Scope for the op.
  """

  def _get_valid_tensor_names(tensor_indices: list[int]) -> list[str]:
    """Gets names of tensors for valid indices."""
    names = []
    for idx in tensor_indices:
      if idx != -1:
        names.append(get_tensor_name(subgraph_tensors[idx]))
    return [name for name in names if name]

  # Op scope is determined by output tensors.
  tensor_names = _get_valid_tensor_names(op.outputs)
  # If no output tensors, use input tensors.
  if not tensor_names:
    tensor_names = _get_valid_tensor_names(op.inputs)

  scope = ";".join(tensor_names)
  # Add a trailing semicolon to help distinguish scopes with the same prefix.
  if tensor_names:
    scope += ";"
  if len(scope) > max_length:
    logging.warning(
        "Op scope is too long, truncating to %d characters. Truncated scope:"
        " %s",
        max_length,
        scope,
    )
  return scope[:max_length]



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Recovers quantized weights from dequantized weights (often from QAT)."""

import dataclasses
from typing import Any, Optional
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import naive_min_max_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils

ALGORITHM_KEY = "dequantized_weight_recovery"
_TFLOpName = qtyping.TFLOperationName
_QuantTransformation = qtyping.QuantTransformation
_IntType = uniform_quantize_tensor.IntType


def _validate_recovered_weights(
    original_vals: np.ndarray,
    quant_vals: np.ndarray,
    scale: np.ndarray,
    tol: float = 1e-4,
):
  """Validates if recovered weights (from the quantized values) are close enough to the original ones.

  Args:
    original_vals: Original values before quantization.
    quant_vals: Quantized values.
    scale: Scale used for quantization.
    tol: Tolerance for the difference between original and recovered values.

  Raises:
    RuntimeError: If the maximum difference between original and recovered
    values exceeds the tolerance.
  """
  recovered_vals = quant_vals * scale
  diff = np.abs(recovered_vals - original_vals).flatten()
  max_diff = diff.max()
  if max_diff > tol:
    raise RuntimeError(
        "Failed to recover the original quantized values from dequantized"
        f" values. Max diff between recovered and original values: {max_diff}"
    )


def _get_scale(arr: np.ndarray, min_scale: float) -> float:
  """Helper function to calculate scale from a 1D array."""
  # Make sure the array includes zero (symmetric quantization).
  arr = np.append(arr, 0)
  unique_vals = np.unique(arr)
  if unique_vals.size > 1:
    diffs = np.diff(unique_vals)
    return float(
        np.maximum(np.min(diffs), min_scale)
    )  # Cast to float to ensure return type consistency
  return min_scale


def get_zp_scale_from_dequantized_symmetric_weights(
    dequant_vals: np.ndarray,
    quantized_dimension: Optional[int] = None,
    min_scale: float = 1e-9,
) -> tuple[np.ndarray, np.ndarray]:
  """Calculates scale and zero point from dequantized and symmetric weights.

  Handles both per-tensor and per-channel (axis) quantization.

  Args:
      dequant_vals: The dequantized weight values (numpy array).
      quantized_dimension:  The dimension along which quantization was performed
        (0 or 1), or None for per-tensor quantization.
      min_scale: The minimum allowed scale value.

  Returns:
      A tuple containing:
          - zero_points: Zero points (all zeros for symmetric quantization).
          - scales: Scales (scalar for per-tensor, array for per-channel).

  Raises:
      ValueError: If `quantized_dimension` is not 0, 1, or None.
  """

  if quantized_dimension not in (0, 1, None):
    raise ValueError(
        f"quantized_dimension must be 0, 1, or None. Got {quantized_dimension}"
    )

  # Use absolute values for symmetric quantization.
  dequant_vals = np.abs(dequant_vals)

  if quantized_dimension is None:
    # Per-tensor quantization: One scale for the entire tensor.
    scales = _get_scale(dequant_vals.flatten(), min_scale)
    scales = np.array([[scales]])
  else:
    # Per-channel quantization: A scale for each slice along the dimension.
    # Create a broadcasted array for per-channel scales. It should have the same
    # number of dimensions as the input, with 1 in all dimensions except for the
    # quantized dimension, which retains its original size.
    scales = np.empty(
        tuple(
            [
                1
                if i != quantized_dimension
                else dequant_vals.shape[quantized_dimension]
                for i in range(dequant_vals.ndim)
            ]
        )
    )
    for i in range(dequant_vals.shape[quantized_dimension]):
      slices = [slice(None)] * dequant_vals.ndim
      slices[quantized_dimension] = i
      vec = dequant_vals[tuple(slices)]
      scales[tuple(slices)] = _get_scale(vec, min_scale)

  zero_points = np.zeros_like(scales, dtype=np.int32)
  return zero_points, scales


def get_tensor_quant_params(
    op_info: qtyping.OpInfo,
    tensor_quant_config: qtyping.TensorQuantizationConfig,
    tensor_content: Optional[np.ndarray] = None,
    tensor_qsv: Optional[dict[str, Any]] = None,
) -> qtyping.UniformQuantParams:
  """Get the quantization parameters for a tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_quant_config: The quantization config for the tensor.
    tensor_content: The content of the tensor.
    tensor_qsv: A dictionary containing the min/max of the tensor.

  Returns:
    The quantization parameters for the tensor.

  Raises:
    ValueError: If the quantization granularity is blockwise, or if the tensor
    is not a symmetric weight tensor.
  """
  # Fallback to naive_min_max_quantize.py for non-weight tensors.
  if tensor_content is None:
    return naive_min_max_quantize.get_tensor_quant_params(
        op_info, tensor_quant_config, tensor_content, tensor_qsv
    )

  if uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity):
    raise ValueError(
        "Blockwise quantization is not supported for dequantized weight"
        " recovery."
    )
  if not tensor_quant_config.symmetric:
    raise ValueError(
        "Only symmetric weights are supported for dequantized weight recovery."
    )

  quantized_dim = common_utils.get_weight_quantized_dim(
      op_info, tensor_content, tensor_quant_config.granularity
  )

  zp, scale = get_zp_scale_from_dequantized_symmetric_weights(
      dequant_vals=tensor_content,
      quantized_dimension=quantized_dim,
  )
  quant_params = qtyping.UniformQuantParams(
      scale=scale,
      zero_point=zp,
      num_bits=tensor_quant_config.num_bits,
      symmetric=tensor_quant_config.symmetric,
      quantized_dimension=quantized_dim,
  )
  quantized_vars = uniform_quantize_tensor.uniform_quantize(
      tensor_content, quant_params
  )
  _validate_recovered_weights(tensor_content, quantized_vars, scale)
  return dataclasses.replace(quant_params, quantized_data=quantized_vars)


def calibrate(
    tfl_op: Any,
    graph_info: qtyping.GraphInfo,
    tensor_content_map: dict[str, np.ndarray],
    inputs_to_ignore: Optional[list[int]] = None,
    outputs_to_ignore: Optional[list[int]] = None,
) -> dict[str, qtyping.QSV]:
  """Collect quantization statistics variable (QSV, e.g., min/max) for the op.

  Args:
    tfl_op: The tfl operation.
    graph_info: Graph information needed to perform quantization for the op.
    tensor_content_map: A map of tensor name to tensor content.
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.

  Returns:
    A dictionary with key as tensor name and value as the collected QSV.
  """
  # Reuse the min/max calibration algorithm from naive_min_max_quantize.py since
  # only weights need to be handled differently.
  return naive_min_max_quantize.min_max_calibrate(
      tfl_op,
      graph_info,
      tensor_content_map,
      inputs_to_ignore,
      outputs_to_ignore,
  )


def init_qsvs(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    inputs_to_ignore: Optional[list[int]] = None,
    outputs_to_ignore: Optional[list[int]] = None,
) -> qtyping.QSV:
  """Initialize the QSVs.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.

  Returns:
    QSVs.
  """
  # Reuse the min/max calibration algorithm from naive_min_max_quantize.py since
  # only weights need to be handeled differently.
  return naive_min_max_quantize.init_qsvs(
      op_info, graph_info, inputs_to_ignore, outputs_to_ignore
  )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import pathlib
from typing import cast

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import mse
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils


class MseQuantizeTest(parameterized.TestCase):
  """Tests for general functions for MSE."""

  def setUp(self):
    super().setUp()
    np.random.seed(666)
    self._test_model_path = str(
        pathlib.Path(test_utils.get_path_to_datafile("../../tests/models"))
        / "conv_fc_mnist.tflite",
    )
    self._test_model = tfl_flatbuffer_utils.read_model(self._test_model_path)
    # The test model has one subgraph for now.
    self._graph_info = qtyping.GraphInfo(
        subgraph_tensors=self._test_model.subgraphs[0].tensors,
        buffers=self._test_model.buffers,
    )
    self._tensor_name_to_qsv = {}
    subgraph0 = self._test_model.subgraphs[0]
    self._subgraph_op_index = 3
    self._fc_op = subgraph0.operators[self._subgraph_op_index]
    self._fc_op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        subgraph_op_index=self._subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=None,
        ),
    )

  def test_get_tensor_quant_params_raises_error_with_unsupported_symmetry(self):
    err_msg = "Unsupported symmetry"
    test_data = np.array([[-7, 7], [4, -4], [4, -4], [7, 7]])
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: err_msg in str(err)
    ):
      _ = mse.get_tensor_quant_params(
          op_info=self._fc_op_info,
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              symmetric=False,
              granularity=qtyping.QuantGranularity.CHANNELWISE,
          ),
          tensor_content=test_data,
      )

  def test_get_tensor_quant_params_raises_error_with_unsupported_granularity(
      self,
  ):
    err_msg = "Blockwise quantization is not supported"
    test_data = np.array([[-7, 7], [4, -4], [4, -4], [7, 7]])
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: err_msg in str(err)
    ):
      _ = mse.get_tensor_quant_params(
          op_info=self._fc_op_info,
          tensor_quant_config=qtyping.TensorQuantizationConfig(
              num_bits=4,
              symmetric=True,
              granularity=qtyping.QuantGranularity.BLOCKWISE_32,
          ),
          tensor_content=test_data,
      )

  def test_get_tensor_quant_params_succeeds_with_qsv(self):
    # Fall back to naive_min_max_quantize.py for non-weight tensors.
    tensor_quant_params = mse.get_tensor_quant_params(
        op_info=self._fc_op_info,
        tensor_quant_config=qtyping.TensorQuantizationConfig(
            num_bits=8,
            granularity=qtyping.QuantGranularity.TENSORWISE,
        ),
        tensor_qsv={
            "min": np.array([-1]),
            "max": np.array([1]),
        },
    )

    self.assertIsNone(tensor_quant_params.quantized_dimension)
    scale = tensor_quant_params.scale
    self.assertEqual(scale.shape, (1,))
    self.assertSequenceAlmostEqual(scale.flatten(), [1 / 127])

    # Zero point should be zero for symmetric quantization.
    zp = tensor_quant_params.zero_point
    self.assertEqual(np.sum(zp), 0)
    self.assertEqual(zp.shape, (1,))

  def test_get_tensor_quant_params_succeeds_with_tensorwise_granularity(self):
    test_data = np.array([
        [-1e5, 25, -50, 75, -100, 125],
        [25, -30, 50, -75, 1e5, -125],
        [50, -60, 70, -80, 90, -100],
    ])
    tensor_config = qtyping.TensorQuantizationConfig(
        num_bits=4,
        symmetric=True,
        granularity=qtyping.QuantGranularity.TENSORWISE,
    )
    fc_op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        subgraph_op_index=self._subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=tensor_config,
        ),
    )
    quant_params = mse.get_tensor_quant_params(
        op_info=fc_op_info,
        tensor_quant_config=tensor_config,
        tensor_content=test_data,
    )

    with self.subTest(name="CheckQuantParamsShapes"):
      self.assertEqual(quant_params.zero_point.shape, (1, 1))
      self.assertEqual(quant_params.scale.shape, (1, 1))
      self.assertIsNone(quant_params.quantized_dimension)
      self.assertIsNotNone(quant_params.quantized_data)
      self.assertTupleEqual(
          cast(np.ndarray, quant_params.quantized_data).shape, test_data.shape
      )

    with self.subTest(name="CheckQuantParamsValues"):
      self.assertTrue(np.all(quant_params.zero_point == 0))

  def test_get_tensor_quant_params_succeeds_with_channelwise_granularity(self):
    # Test that the call generates quant params that are appropriately shaped,
    # have some clipping, and correct config values without checking the
    # actual values numerically.
    test_data = np.array([
        [-1e5, 25, -50, 75, -100, 125],
        [25, -30, 50, -75, 1e5, -125],
        [50, -60, 70, -80, 90, -100],
    ])
    tensor_config = qtyping.TensorQuantizationConfig(
        num_bits=4,
        symmetric=True,
        granularity=qtyping.QuantGranularity.CHANNELWISE,
    )
    fc_op_info = qtyping.OpInfo(
        op=self._fc_op,
        op_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        subgraph_op_index=self._subgraph_op_index,
        op_quant_config=qtyping.OpQuantizationConfig(
            weight_tensor_config=tensor_config,
        ),
    )
    quant_params = mse.get_tensor_quant_params(
        op_info=fc_op_info,
        tensor_quant_config=tensor_config,
        tensor_content=test_data,
    )

    with self.subTest(name="CheckQuantParamsShapes"):
      self.assertEqual(quant_params.zero_point.shape, (test_data.shape[0], 1))
      self.assertEqual(quant_params.scale.shape, (test_data.shape[0], 1))
      self.assertIsNotNone(quant_params.quantized_data)
      self.assertTupleEqual(
          cast(np.ndarray, quant_params.quantized_data).shape, test_data.shape
      )

    with self.subTest(name="CheckQuantParamsValues"):
      self.assertTrue(np.all(quant_params.zero_point == 0))
      self.assertEqual(quant_params.quantized_dimension, 0)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from absl.testing import parameterized
import absl.testing.absltest as absltest
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import constrained_ops_utils


_OpQuantConstraint = common_utils.OpQuantConstraint


class ConstrainedOpsUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="same_as_input_scale",
          constraint=_OpQuantConstraint.SAME_AS_INPUT_SCALE,
          expected_num_ops=18,
      ),
      dict(
          testcase_name="same_as_output_scale",
          constraint=_OpQuantConstraint.SAME_AS_OUTPUT_SCALE,
          expected_num_ops=7,
      ),
      dict(
          testcase_name="no_constrain",
          constraint=_OpQuantConstraint.NO_CONSTRAIN,
          expected_num_ops=25,
      ),
  )
  def test_get_constrained_op_list(self, constraint, expected_num_ops):
    constrained_ops = constrained_ops_utils.get_constrained_op_list(constraint)
    self.assertLen(constrained_ops, expected_num_ops)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Performs naive min/max uniform quantization."""

import dataclasses
from typing import Any, Optional
import numpy as np
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.algorithms.uniform_quantize import common_quantize
from ai_edge_quantizer.algorithms.uniform_quantize import uniform_quantize_tensor
from ai_edge_quantizer.algorithms.utils import common_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils

ALGORITHM_KEY = "min_max_uniform_quantize"
_TFLOpName = qtyping.TFLOperationName
_QuantTransformation = qtyping.QuantTransformation
_IntType = uniform_quantize_tensor.IntType


def get_tensor_quant_params(
    op_info: qtyping.OpInfo,
    tensor_quant_config: qtyping.TensorQuantizationConfig,
    tensor_content: Optional[np.ndarray] = None,
    tensor_qsv: Optional[dict[str, Any]] = None,
) -> qtyping.UniformQuantParams:
  """Get the quantization parameters for a tensor.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    tensor_quant_config: The quantization config for the tensor.
    tensor_content: The content of the tensor.
    tensor_qsv: A dictionary containingthe min/max of the tensor.

  Returns:
    The quantization parameters for the tensor.
  """
  # Get quant params.
  if tensor_qsv is None:
    if tensor_content is not None:
      # We need min/max to calculate quantization parameters, which
      # should be collected during the calibration process. However,
      # weight-only and DRQ do not require calibration, thus it is
      # possible that this information is missing here. In that case we
      # collect min/max on the spot.
      tensor_min_max = common_quantize.init_tensor_min_max(
          tensor_content,
          op_info,
      )
    else:
      raise ValueError(
          f"{op_info.op_name}(index: {op_info.subgraph_op_index}) not found in"
          " tensor_name_to_qsv. Check if the correct calibration results are"
          " passed into the ParamsGenerator."
      )
  else:
    tensor_min_max = tensor_qsv

  if "min" not in tensor_min_max or "max" not in tensor_min_max:
    raise ValueError(
        "min and max must be provided to produce tensor quantization"
        " parameters. Check if the correct calibration results are passed into"
        " the ParamsGenerator."
    )
  clipping_values = None
  zp, scale = uniform_quantize_tensor.tensor_zp_scale_from_min_max(
      tensor_min_max["min"],
      tensor_min_max["max"],
      tensor_quant_config.num_bits,
      tensor_quant_config.symmetric,
      tensor_quant_config.granularity,
      clipping_values,
  )
  quantized_dim = common_utils.get_weight_quantized_dim(
      op_info, tensor_content, tensor_quant_config.granularity
  )
  quant_params = qtyping.UniformQuantParams(
      scale=scale,
      zero_point=zp,
      num_bits=tensor_quant_config.num_bits,
      symmetric=tensor_quant_config.symmetric,
      quantized_dimension=quantized_dim,
      block_size=uniform_quantize_tensor.extract_block_size_from_granularity(
          tensor_quant_config.granularity
      ),
  )
  if tensor_content is None:
    return quant_params

  quantized_vars = uniform_quantize_tensor.uniform_quantize(
      tensor_content,
      quant_params,
      uniform_quantize_tensor.is_blockwise(tensor_quant_config.granularity),
  )
  # Update with quantized values.
  return dataclasses.replace(quant_params, quantized_data=quantized_vars)


def check_if_quantized(tensor: Any) -> bool:
  """Checks if the tensor is quantized."""
  return (
      tensor.quantization is not None and tensor.quantization.scale is not None
  )


# TODO: b/333731147 - Use named tuple to store min/max.
def init_qsvs(
    op_info: qtyping.OpInfo,
    graph_info: qtyping.GraphInfo,
    inputs_to_ignore: Optional[list[int]] = None,
    outputs_to_ignore: Optional[list[int]] = None,
) -> qtyping.QSV:
  """Initialize the QSVs.

  Args:
    op_info: Aggregated information about the op (e.g., quantization config).
    graph_info: Graph information needed to perform quantization for the op.
    inputs_to_ignore: Operand indices to ignore.
    outputs_to_ignore: Result indices to ignore.

  Returns:
    QSVs.
  """
  op_qsvs = {}

  inputs_to_ignore = inputs_to_ignore or []
  quantized_inputs_to_ignore = [
      opr_idx
      for opr_idx, tensor_idx in enumerate(op_info.op.inputs)
      if check_if_quantized(graph_info.subgraph_tensors[tensor_idx])
  ]
  inputs_to_ignore.extend(quantized_inputs_to_ignore)

  outputs_to_ignore = outputs_to_ignore or []
  for opr_idx, tensor_idx in enumerate(op_info.op.inputs):
    if tensor_idx != -1 and opr_idx not in inputs_to_ignore:
      tensor = graph_info.subgraph_tensors[tensor_idx]
      tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
      tensor_data = tfl_flatbuffer_utils.get_tensor_data(
          tensor, graph_info.buffers
      )
      op_qsvs[tensor_name] = common_quantize.init_tensor_min_max(
          tensor_data,
          op_info,
      )
  for res_idx, tensor_idx in enumerate(op_info.op.outputs):
    if tensor_idx != -1 and res_idx not in outputs_to_ignore:
      tensor = graph_info.subgraph_tensors[tensor_idx]
      tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
      tensor_data = tfl_flatbuffer_utils.get_tensor_data(
          tensor, graph_info.buffers
      )
      op_qsvs[tensor_name] = common_quantize.init_tensor_min_max(
          tensor_data,
          op_info,
      )
  return op_qsvs


def min_max_calibrate(
    tfl_op: Any,
    graph_info: qtyping.GraphInfo,
    tensor_content_map: dict[str, np.ndarray],
    inputs_to_ignore: Optional[list[int]] = None,
    outputs_to_ignore: Optional[list[int]] = None,
    valid_range: tuple[float, float] = (-3e38, 3e38),
) -> dict[str, qtyping.QSV]:
  """Collect quantization statistics variable (QSV, e.g., min/max) for the op.

  Args:
    tfl_op: The tfl operation.
    graph_info: Graph information needed to perform quantization for the op.
    tensor_content_map: A map of tensor name to tensor content.
    inputs_to_ignore: Input tensor indices to ignore.
    outputs_to_ignore: Output tensor indices to ignore.
    valid_range: The valid range for tensor content, excluding the boundaries.
      Tensor values outside this range are ignored during calibration. Defaults
      to an approximate bfloat16 range. This range is chosen to address issues
      with `padv2` where a bfloat16 -inf padding constant can cause problems.
      Values exceeding this range can lead to quantization issues and are
      therefore excluded from min/max calibration.

  Returns:
    A dictionary with key as tensor name and value as the collected QSV.
  """
  op_qsvs = {}
  min_val, max_val = valid_range

  def _collect_activation_tensor_min_max(tensor_idx):
    tensor = graph_info.subgraph_tensors[tensor_idx]
    tensor_data = tfl_flatbuffer_utils.get_tensor_data(
        tensor, graph_info.buffers
    )
    # Skip constant tensors.
    if tensor_data is not None:
      return
    tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
    tensor_content = tensor_content_map[tensor_name]
    qsv_shape = (1,) * tensor_content.ndim
    filter_mask = (tensor_content > min_val) & (tensor_content < max_val)
    if np.any(filter_mask):
      tensor_content = tensor_content[filter_mask]
    # Reshape is needed to ensure the scalar min/max have the same number of
    # dimensions as the input tensor array, for compatibility with subsequent
    # operations.
    op_qsvs[tensor_name] = {
        "min": np.min(tensor_content, axis=None).reshape(qsv_shape),
        "max": np.max(tensor_content, axis=None).reshape(qsv_shape),
    }

  inputs_to_ignore = inputs_to_ignore or []
  quantized_inputs_to_ignore = [
      opr_idx
      for opr_idx, tensor_idx in enumerate(tfl_op.inputs)
      if check_if_quantized(graph_info.subgraph_tensors[tensor_idx])
  ]
  inputs_to_ignore.extend(quantized_inputs_to_ignore)
  outputs_to_ignore = outputs_to_ignore or []
  for i, tensor_idx in enumerate(tfl_op.inputs):
    if tensor_idx != -1 and i not in inputs_to_ignore:
      _collect_activation_tensor_min_max(tensor_idx)
  for i, tensor_idx in enumerate(tfl_op.outputs):
    if tensor_idx != -1 and i not in outputs_to_ignore:
      _collect_activation_tensor_min_max(tensor_idx)

  return op_qsvs



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Manages model quantization recipe (configuration) for the quantizer."""

import collections
import dataclasses
import re
from typing import Any, Optional
from absl import logging
from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import qtyping

# A collection of quantization configuration.
# Key: scope regex.
# Value: list of OpQuantizationRecipe in dictionary format.
ModelQuantizationRecipe = list[dict[str, Any]]
# Expose algorithm names to users.
AlgorithmName = algorithm_manager.AlgorithmName

_TFLOpName = qtyping.TFLOperationName
_OpQuantizationConfig = qtyping.OpQuantizationConfig
_TensorQuantizationConfig = qtyping.TensorQuantizationConfig


@dataclasses.dataclass
class OpQuantizationRecipe:
  """Dataclass for quantization configuration under a scope.

  This class is the main entry point to a recipe schema. There could be a single
  instance of this associated with a model (with `regex=.*` if the full model is
  to be quantized with the same spec), or multiple instances targeting different
  `regex` or `operation`.

  Attributes:
    regex: Regular expression for scope name matching. Any op that matches
      `regex` will be quantized according to this instance. The narrowest scope
      would be the full output tensor name of an op. The widest scope would be
      '.*' which applies to the full model.
    operation: Target TFL operation. * for any supported TFLite operation.
    algorithm_key: Algorithm key to be applied. This can be any one of the
      strings as enumerated in `AlgorithmName`.
    op_config: Quantization configuration to be applied for the op.
  """

  regex: str
  operation: _TFLOpName
  algorithm_key: str
  op_config: _OpQuantizationConfig = dataclasses.field(
      default_factory=_OpQuantizationConfig
  )


class RecipeManager:
  """Sets the quantization recipe for target model.

  This class is internal to Quantizer to help manage loading recipes and
  resolving conflicts between recipes input by the user.
  """

  def __init__(self):
    """Scope name config.

    Key: scope regex. ".*" for all scopes.
    Value: list of operator quantization settings under the scope.
    The priority between rules are determined by the order they entered: later
    one has higher priority.
    """
    self._scope_configs: collections.OrderedDict[
        str, list[OpQuantizationRecipe]
    ] = collections.OrderedDict()

  def add_quantization_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      op_config: Optional[_OpQuantizationConfig] = None,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ) -> None:
    """Adds a quantization configuration.

    Conflict arises when we are trying to set an operation under a certain regex
    which is already existed in the config dictionary. Under such circumstance,
    the new config is used to replace the previous one.

    We also have special treatment for _TFLOperationKey.ALL. If the new config
    is on _TFLOperationKey.ALL and there are existing op configs inside the same
    scope, we clear the previous configs and use _TFLOperationKey.ALL.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation. * for all supported TFLite
        operation.
      op_config: Quantization configuration which will be used to update the
        default configuration. None or empty dict means the default
        configuration will be used.
      algorithm_key: Algorithm key to be applied.
    """
    try:
      algorithm_manager.AlgorithmName(algorithm_key)
    except ValueError as e:
      raise ValueError(f'Unsupported algorithm key: {algorithm_key}.') from e

    if op_config is None:
      op_config = _OpQuantizationConfig()

    config = OpQuantizationRecipe(
        regex, operation_name, algorithm_key, op_config
    )
    # Special care if trying to set all ops to some config.
    if config.operation == _TFLOpName.ALL_SUPPORTED:
      self._scope_configs[regex] = [config]
      return

    if algorithm_key != AlgorithmName.NO_QUANTIZE:
      algorithm_manager.check_op_quantization_config(
          algorithm_key, operation_name, op_config
      )

    if regex not in self._scope_configs:
      self._scope_configs[regex] = [config]
    else:
      # Reiterate configs to avoid duplication on op settings.
      configs = []
      is_new_op = True
      for existing_config in self._scope_configs[regex]:
        if existing_config.operation == config.operation:
          is_new_op = False
          op_config = config
          logging.warning(
              'Overwrite operation %s config under scope_regex %s with %s.',
              existing_config.operation,
              regex,
              config,
          )
        else:
          op_config = existing_config
        configs.append(op_config)
      if is_new_op:
        configs.append(config)
      self._scope_configs[regex] = configs

  def get_quantization_configs(
      self,
      target_op_name: _TFLOpName,
      scope_name: str,
  ) -> tuple[str, _OpQuantizationConfig]:
    """Gets the algorithm key and quantization configuration for an op.

    We respect the latest valid config and fall back to no quantization.
    Specifically, we search the quantization configuration in the order of the
    scope configs. If there are two or more matching settings, the latest one
    will be used.


    Args:
      target_op_name: Target TFLite operation. * for all supported TFLite
        operation.
      scope_name: Name of the target scope.

    Returns:
       A tuple of quantization algorithm, and quantization configuration.
    """
    result_key, result_config = (
        AlgorithmName.NO_QUANTIZE,
        _OpQuantizationConfig(),
    )
    for scope_regex, recipes in self._scope_configs.items():
      if re.search(scope_regex, scope_name):
        for recipe in recipes:
          if (
              recipe.operation != _TFLOpName.ALL_SUPPORTED
              and recipe.operation != target_op_name
          ):
            continue
          selected_recipe = recipe
          if selected_recipe.algorithm_key != AlgorithmName.NO_QUANTIZE:
            # The selected recipe must contain a supported config.
            try:
              algorithm_manager.check_op_quantization_config(
                  recipe.algorithm_key, target_op_name, recipe.op_config
              )
            except ValueError:
              continue  # Skip the recipe if it is not supported.
          result_config = selected_recipe.op_config
          result_key = selected_recipe.algorithm_key

    return result_key, result_config

  def get_quantization_recipe(self) -> ModelQuantizationRecipe:
    """Gets the full quantization recipe from the manager.

    Returns:
      A list of quantization configs in the recipe.
    """
    ret = []
    for _, scope_config in self._scope_configs.items():
      for quant_config in scope_config:
        config = dict()
        config['regex'] = quant_config.regex
        config['operation'] = quant_config.operation
        config['algorithm_key'] = quant_config.algorithm_key
        config['op_config'] = quant_config.op_config.to_dict()
        ret.append(config)
    return ret

  def load_quantization_recipe(
      self, quantization_recipe: ModelQuantizationRecipe
  ) -> None:
    """Loads the quantization recipe to the manager.

    Args:
      quantization_recipe: A configuration dictionary which is generated by
        get_full_config.
    """
    self._scope_configs = collections.OrderedDict()
    for config in quantization_recipe:
      self.add_quantization_config(
          config['regex'],
          config['operation'],
          _OpQuantizationConfig.from_dict(config['op_config'])
          if config['algorithm_key'] != AlgorithmName.NO_QUANTIZE
          else None,
          config['algorithm_key'],
      )

  def need_calibration(self) -> bool:
    """Check if the recipe requires calibration."""
    # At the moment, only SRQ requires calibration.
    for op_quant_config in self.get_quantization_recipe():
      if (
          op_quant_config['op_config']['compute_precision']
          == qtyping.ComputePrecision.INTEGER
          and 'activation_tensor_config' in op_quant_config['op_config']
      ):
        return True
    return False

  def add_dynamic_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      num_bits: int,
      granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a dynamic quantization configuration to the recipe.

    During dynamic quantization, activations are not processed by AEQ and
    remain in float format. The runtime kernel is expected to quantize these
    activations on-the-fly, as indicated by compute_precision=Integer and
    explicit_dequantize=False.

    The model quality may suffer due to the on-the-fly quantization. If quality
    is a concern, consider using weight-only
    quantization.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation.
      num_bits: Number of bits for quantization.
      granularity: Granularity of quantization.
      algorithm_key: Algorithm key to be applied.
    """
    weight_config = qtyping.TensorQuantizationConfig(
        num_bits=num_bits,
        symmetric=True,  # LiteRT kernels only support symmetric quantized
        # weights.
        granularity=granularity,
    )
    self.add_quantization_config(
        regex,
        operation_name,
        op_config=_OpQuantizationConfig(
            weight_tensor_config=weight_config,
            compute_precision=qtyping.ComputePrecision.INTEGER,
            explicit_dequantize=False,
        ),
        algorithm_key=algorithm_key,
    )

  def add_weight_only_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      num_bits: int,
      granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a weight only quantization configuration to the recipe.

    In weight-only quantization, weights are quantized, but the actual operation
    (op) computation remains in float. The quantized weight is explicitly
    dequantized before being fed into the op. This is achieved by inserting a
    dequantize op between the quantized weight and the consuming op. To enable
    this, both compute_precision will be set to Float and explicit_dequantize to
    True.

    Weight-only quantization is useful for reducing model size but may
    not decrease latency due to float computation. However, quantized model
    generally has better quality than other quantization options (e.g., dynamic
    range quantization) due to no loss of precision on activations. If latency
    is a concern, consider using dynamic quantization.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation.
      num_bits: Number of bits for quantization.
      granularity: Granularity of quantization.
      algorithm_key: Algorithm key to be applied.
    """
    # Default to integer quantization but allow float quantization for
    # FLOAT_CASTING algorithm. This is to support weight-only quantization with
    # fp16 weights.
    weight_dtype = qtyping.TensorDataType.INT
    if algorithm_key == AlgorithmName.FLOAT_CASTING:
      weight_dtype = qtyping.TensorDataType.FLOAT

    weight_config = qtyping.TensorQuantizationConfig(
        num_bits=num_bits,
        symmetric=True,  # TFL kernels only support symmetric quantized weights.
        granularity=granularity,
        dtype=weight_dtype,
    )
    self.add_quantization_config(
        regex,
        operation_name,
        op_config=_OpQuantizationConfig(
            weight_tensor_config=weight_config,
            compute_precision=qtyping.ComputePrecision.FLOAT,
            explicit_dequantize=True,
        ),
        algorithm_key=algorithm_key,
    )

  def add_static_config(
      self,
      regex: str,
      operation_name: _TFLOpName,
      activation_num_bits: int,
      weight_num_bits: int,
      weight_granularity: qtyping.QuantGranularity = qtyping.QuantGranularity.CHANNELWISE,
      algorithm_key: str = algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
  ):
    """Adds a static range quantization configuration to the recipe.

    In static quantization, both weights and activations are quantized. This
    requires a calibration step to determine the quantization parameters (e.g.,
    min/max ranges) for activations. The quantized model uses integer arithmetic
    for computations, which can lead to significant latency reductions.

    However, calibration is needed to determine the quantization parameters for
    activations, which requires sample data and may lead to quality loss. If
    there is no hardware requirement for full integer quantization, consider
    using dynamic quantization for simplicity.

    Args:
      regex: Regular expression for layer name matching.
      operation_name: Target TFLite operation.
      activation_num_bits: Number of bits for activation quantization.
      weight_num_bits: Number of bits for weight quantization.
      weight_granularity: Granularity of weight quantization.
      algorithm_key: Algorithm key to be applied.
    """
    if activation_num_bits not in [16, 8]:
      raise ValueError(
          'Activation quantization is only supported for 16 or 8 bits.'
      )
    # INT16 is symmetric and INT8 is asymmetric due to LiteRT kernel
    # limitations.
    activation_symmetric = activation_num_bits == 16
    activation_config = qtyping.TensorQuantizationConfig(
        num_bits=activation_num_bits, symmetric=activation_symmetric
    )
    weight_config = qtyping.TensorQuantizationConfig(
        num_bits=weight_num_bits,
        symmetric=True,  # TFL kernels only support symmetric quantized weights.
        granularity=weight_granularity,
    )
    self.add_quantization_config(
        regex,
        operation_name,
        op_config=_OpQuantizationConfig(
            activation_tensor_config=activation_config,
            weight_tensor_config=weight_config,
            compute_precision=qtyping.ComputePrecision.INTEGER,
            explicit_dequantize=False,
        ),
        algorithm_key=algorithm_key,
    )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""E2E tests for the quantizer for model with embedding_lookup."""

from absl.testing import parameterized
import absl.testing.absltest as absltest
import numpy as np

from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils

_ComputePrecision = qtyping.ComputePrecision
_OpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_OpQuantConfig = qtyping.OpQuantizationConfig

_RNG = np.random.default_rng(66)


def _get_dummy_data(num_samples):
  data = []
  for _ in range(num_samples):
    data.append({'inputs': _RNG.uniform(size=(1,)).astype(np.int32)})
  return data


# def _get_test_data(num_samples: int = 8):
#   return _get_dummy_data(num_samples)


class EmbeddingLookupTest(test_utils.BaseOpTestCase):

  def setUp(self):
    super().setUp()
    self.float_model_path = test_utils.get_path_to_datafile(
        '../models/embedding_lookup.tflite'
    )
    self._quantizer = quantizer.Quantizer(self.float_model_path)

  @parameterized.parameters(
      ('../../recipes/default_af32w8float_recipe.json', 1700),
      ('../../recipes/default_af32w4float_recipe.json', 1600),
      ('../../recipes/dynamic_legacy_wi8_afp32_recipe.json', 1400),
      ('../../recipes/dynamic_wi8_afp32_recipe.json', 1400),
  )
  def test_embedding_lookup_model_int_weight_only(
      self, recipe_path, expected_model_size
  ):
    recipe_path = test_utils.get_path_to_datafile(recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertFalse(self._quantizer.need_calibration)
    quant_result = self._quantizer.quantize()
    # Check model size.
    self.assertLess(len(quant_result.quantized_model), expected_model_size)

  @parameterized.product(
      weight_bit_width=[4, 8],
      quantization_method=['add_weight_only_config', 'add_dynamic_config'],
  )
  def test_embedding_lookup_model_mse_quantization(
      self, weight_bit_width, quantization_method
  ):
    config_setter = getattr(self._quantizer, quantization_method)
    config_setter(
        regex='.*',
        operation_name=_OpName.EMBEDDING_LOOKUP,
        num_bits=weight_bit_width,
        algorithm_key=quantizer.AlgorithmName.MSE,
    )
    self.assertFalse(self._quantizer.need_calibration)
    quant_result = self._quantizer.quantize()
    # Only check model size for now.
    self.assertLess(len(quant_result.quantized_model), 1700)

  def test_embedding_lookup_model_fp16_weight_only(self):
    self._quantizer.update_quantization_recipe(
        regex='.*',
        algorithm_key=quantizer.AlgorithmName.FLOAT_CASTING,
        operation_name=_OpName.EMBEDDING_LOOKUP,
        op_config=_OpQuantConfig(
            weight_tensor_config=_TensorQuantConfig(
                num_bits=16, dtype=qtyping.TensorDataType.FLOAT
            ),
            compute_precision=_ComputePrecision.FLOAT,  # WEIGHT_ONLY
            explicit_dequantize=True,
        ),
    )
    quant_result = self._quantizer.quantize()
    self.assertLess(len(quant_result.quantized_model), 2000)

    # TODO: b/364405203 - Enable after 0 signature works.
    # comparion_result = self._quantizer.validate(
    #     error_metrics='mse',
    #     signature_test_data=_get_test_data(),
    # )
    # self._check_comparion_result(
    #     comparion_result,
    #     weight_tolerance=1e-5,
    #     output_tolerance=1e-5,
    # )

  @parameterized.product(weight_bit_width=[4, 8])
  def test_hadamard_rotation_accuracy_and_size_within_tolerance(
      self, weight_bit_width
  ):
    self._quantizer.load_quantization_recipe([
        {
            'regex': '.*',
            'operation': qtyping.TFLOperationName.EMBEDDING_LOOKUP,
            'algorithm_key': quantizer.AlgorithmName.HADAMARD_ROTATION,
            'op_config': {
                'weight_tensor_config': {
                    'dtype': qtyping.TensorDataType.INT,
                    'num_bits': weight_bit_width,
                    'granularity': qtyping.QuantGranularity.CHANNELWISE,
                },
                'compute_precision': qtyping.ComputePrecision.FLOAT,
                'explicit_dequantize': True,
            },
        },
    ])
    quant_result = self._quantizer.quantize()
    with self.subTest(name='ModelSizeReduction'):
      self.assertLess(len(quant_result.quantized_model), 1650)

    validation_result = self._quantizer.validate(
        test_data={
            'lookup': [
                {'lookup': np.random.randint(0, 15, size=(1,), dtype=np.int32)}
            ],
        },
        error_metrics='mse',
    )
    with self.subTest(name='OutputErrors'):
      self.assert_output_errors_below_tolerance(validation_result, 1e-5)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import json
import pathlib

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

import os
import io
from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import default_policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils

_ComputePrecision = qtyping.ComputePrecision
_TFLOpName = qtyping.TFLOperationName
_TensorQuantConfig = qtyping.TensorQuantizationConfig
_TensorDataType = qtyping.TensorDataType
_AlgorithmName = quantizer.AlgorithmName

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile('')
_MULTI_SIGNATURE_CALIBRATION_DATASET = {
    'add': [{'x': np.array([2.0], dtype=np.float32)}],
    'multiply': [{'x': np.array([1.0], dtype=np.float32)}],
}
_RNG = np.random.default_rng(66)


def _get_calibration_data(num_samples: int = 16):
  calibration_samples = []
  for _ in range(num_samples):
    calibration_samples.append(
        {'conv2d_input': _RNG.uniform(size=(1, 28, 28, 1)).astype(np.float32)}
    )
  calibration_data = {
      tfl_interpreter_utils.DEFAULT_SIGNATURE_KEY: calibration_samples,
  }
  return calibration_data


def _is_all_signature_defs_inputs_float(model_content: bytes):
  tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(model_content)
  for signature_key in tfl_interpreter.get_signature_list():
    input_details = tfl_interpreter.get_signature_runner(
        signature_key
    ).get_input_details()
    for tensor_details in input_details.values():
      if tensor_details['dtype'] != np.float32:
        return False
  return True


def _is_all_signature_defs_outputs_float(model_content: bytes):
  tfl_interpreter = tfl_interpreter_utils.create_tfl_interpreter(model_content)
  for signature_key in tfl_interpreter.get_signature_list():
    output_details = tfl_interpreter.get_signature_runner(
        signature_key
    ).get_output_details()
    for tensor_details in output_details.values():
      if tensor_details['dtype'] != np.float32:
        return False
  return True


class QuantizerTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._tmp_save_path = self.create_tempdir().full_path
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite'
    )
    self._test_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/default_af32w8float_recipe.json',
    )
    with open(self._test_recipe_path) as json_file:
      self._test_recipe = json.load(json_file)
    self._quantizer = quantizer.Quantizer(
        self._test_model_path, self._test_recipe_path
    )

  def test_update_quantization_recipe_succeeds(self):
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    scope_regex = '.*/Dense/.*'
    new_op_config = qtyping.OpQuantizationConfig(
        weight_tensor_config=_TensorQuantConfig(num_bits=4, symmetric=True),
        compute_precision=_ComputePrecision.INTEGER,  # DRQ.
    )
    self._quantizer.update_quantization_recipe(
        regex=scope_regex,
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=new_op_config,
    )
    updated_recipe = self._quantizer.get_quantization_recipe()
    self.assertLen(updated_recipe, 2)

    added_config = updated_recipe[-1]
    self.assertEqual(added_config['regex'], scope_regex)
    self.assertEqual(
        added_config['op_config']['compute_precision'],
        new_op_config.compute_precision,
    )

  def test_add_dynamic_config_succeeds(self):
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    scope_regex = '.*/Dense/.*'
    self._quantizer.add_dynamic_config(
        regex=scope_regex,
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        num_bits=8,
    )
    updated_recipe = self._quantizer.get_quantization_recipe()
    self.assertLen(updated_recipe, 2)

    added_config = updated_recipe[-1]
    self.assertEqual(added_config['regex'], scope_regex)
    self.assertEqual(
        added_config['op_config']['compute_precision'],
        qtyping.ComputePrecision.INTEGER,
    )
    self.assertFalse(added_config['op_config']['explicit_dequantize'])
    self.assertEqual(
        added_config['op_config']['weight_tensor_config']['num_bits'], 8
    )

  def test_add_weight_only_config_succeeds(self):
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    scope_regex = '.*/Dense/.*'
    self._quantizer.add_weight_only_config(
        regex=scope_regex,
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        num_bits=4,
    )
    updated_recipe = self._quantizer.get_quantization_recipe()
    self.assertLen(updated_recipe, 2)

    added_config = updated_recipe[-1]
    self.assertEqual(added_config['regex'], scope_regex)
    self.assertEqual(
        added_config['op_config']['compute_precision'],
        qtyping.ComputePrecision.FLOAT,
    )
    self.assertTrue(added_config['op_config']['explicit_dequantize'])
    self.assertEqual(
        added_config['op_config']['weight_tensor_config']['num_bits'], 4
    )

  def test_add_static_config_succeeds(self):
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    scope_regex = '.*/Dense/.*'
    self._quantizer.add_static_config(
        regex=scope_regex,
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        activation_num_bits=8,
        weight_num_bits=4,
    )
    updated_recipe = self._quantizer.get_quantization_recipe()
    self.assertLen(updated_recipe, 2)

    added_config = updated_recipe[-1]
    self.assertEqual(added_config['regex'], scope_regex)
    self.assertEqual(
        added_config['op_config']['compute_precision'],
        qtyping.ComputePrecision.INTEGER,
    )
    self.assertFalse(added_config['op_config']['explicit_dequantize'])
    self.assertEqual(
        added_config['op_config']['activation_tensor_config']['num_bits'], 8
    )
    self.assertEqual(
        added_config['op_config']['weight_tensor_config']['num_bits'], 4
    )

  def test_load_quantization_recipe_succeeds(self):
    qt = quantizer.Quantizer(self._test_model_path, None)
    qt.load_quantization_recipe(self._test_recipe_path)
    self.assertEqual(qt.get_quantization_recipe(), self._test_recipe)

    # Load a different recipe.
    new_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/dynamic_wi8_afp32_recipe.json',
    )
    with open(new_recipe_path) as json_file:
      new_recipe = json.load(json_file)
    qt.load_quantization_recipe(new_recipe_path)
    self.assertEqual(qt.get_quantization_recipe(), new_recipe)

  @parameterized.parameters(
      'recipes/default_a8w8_recipe.json',
      'recipes/default_a16w8_recipe.json',
  )
  def test_calibrate_required_recipe_succeeds(self, recipe_path):
    recipe_path = str(pathlib.Path(TEST_DATA_PREFIX_PATH) / recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    # Calibrate with empty state.
    calib_data = _get_calibration_data()
    calibration_result = self._quantizer.calibrate(calib_data)
    self.assertLen(calibration_result, 7)

  @parameterized.parameters(
      'recipes/default_a8w8_recipe.json',
      'recipes/default_a16w8_recipe.json',
  )
  def test_reloaded_calibration_succeeds(self, recipe_path):
    recipe_path = str(pathlib.Path(TEST_DATA_PREFIX_PATH) / recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    calib_data = _get_calibration_data()
    calibration_result = self._quantizer.calibrate(calib_data)
    # Load and calibrate again.
    updated_calibration_result = self._quantizer.calibrate(
        calib_data, previous_calibration_result=calibration_result
    )
    self.assertLen(updated_calibration_result, 7)
    self.assertNotEqual(
        calibration_result['StatefulPartitionedCall:0'],
        updated_calibration_result['StatefulPartitionedCall:0'],
    )

  @parameterized.parameters(
      'recipes/dynamic_legacy_wi8_afp32_recipe.json',
      'recipes/dynamic_wi8_afp32_recipe.json',
      'recipes/default_af32w8float_recipe.json',
  )
  def test_calibrate_nonrequired_recipe_succeeds(self, recipe_path):
    recipe_path = str(pathlib.Path(TEST_DATA_PREFIX_PATH) / recipe_path)
    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertFalse(self._quantizer.need_calibration)
    # Empty calibration result if no calibration is required.
    calibration_result = self._quantizer.calibrate(_get_calibration_data())
    self.assertEmpty(calibration_result)

  def test_quantize_no_calibration_succeeds(self):
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    self.assertIsNone(self._quantizer._result.quantized_model)
    quant_result = self._quantizer.quantize()
    self.assertEqual(quant_result.recipe, self._test_recipe)
    self.assertIsNotNone(quant_result.quantized_model)

  @parameterized.parameters(
      'recipes/default_a8w8_recipe.json',
      'recipes/default_a16w8_recipe.json',
  )
  def test_quantize_calibration_needed_succeeds(self, recipe_path):
    recipe_path = str(pathlib.Path(TEST_DATA_PREFIX_PATH) / recipe_path)
    with open(recipe_path) as json_file:
      recipe = json.load(json_file)

    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    calibration_result = self._quantizer.calibrate(_get_calibration_data())

    self.assertIsNone(self._quantizer._result.quantized_model)
    quant_result = self._quantizer.quantize(calibration_result)
    self.assertEqual(quant_result.recipe, recipe)
    self.assertIsNotNone(quant_result.quantized_model)

  @parameterized.parameters(
      'recipes/default_a8w8_recipe.json',
      'recipes/default_a16w8_recipe.json',
  )
  def test_quantize_calibration_needed_raise_error(self, recipe_path):
    recipe_path = str(pathlib.Path(TEST_DATA_PREFIX_PATH) / recipe_path)

    self._quantizer.load_quantization_recipe(recipe_path)
    self.assertTrue(self._quantizer.need_calibration)
    error_message = (
        'Model quantization statistics values (QSVs) are required for the input'
        ' recipe.'
    )
    with self.assertRaisesWithPredicateMatch(
        RuntimeError, lambda err: error_message in str(err)
    ):
      self._quantizer.quantize()

  def test_quantize_no_recipe_raise_error(self):
    qt = quantizer.Quantizer(self._test_model_path, None)
    error_message = 'Can not quantize without a quantization recipe.'
    with self.assertRaisesWithPredicateMatch(
        RuntimeError, lambda err: error_message in str(err)
    ):
      qt.quantize()

  def test_save_succeeds(self):
    model_name = 'test_model'
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    result = self._quantizer.quantize()
    result.save(self._tmp_save_path, model_name)
    saved_recipe_path = str(
        pathlib.Path(self._tmp_save_path) / (model_name + '_recipe.json')
    )
    with open(saved_recipe_path) as json_file:
      saved_recipe = json.load(json_file)
    self.assertEqual(saved_recipe, self._test_recipe)

  def test_saved_legacy_recipe_lacks_block_size(self):
    model_name = 'test_model'
    legacy_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/dynamic_legacy_wi8_afp32_recipe.json',
    )
    self._quantizer.load_quantization_recipe(legacy_recipe_path)
    result = self._quantizer.quantize()
    result.save(self._tmp_save_path, model_name)
    saved_recipe_path = str(
        pathlib.Path(self._tmp_save_path) / (model_name + '_recipe.json')
    )
    with open(saved_recipe_path) as json_file:
      saved_recipe = json.load(json_file)
    with open(legacy_recipe_path) as json_file:
      legacy_recipe = json.load(json_file)

    self.assertNotEqual(saved_recipe, legacy_recipe)

    # Verify that the default test recipe contains 'block_size'.
    has_block_size = False
    for config in legacy_recipe:
      op_config = config.get('op_config')
      if op_config:
        weight_config = op_config.get('weight_tensor_config')
        if weight_config and 'block_size' in weight_config:
          has_block_size = True
          break
    self.assertTrue(has_block_size)

    # Verify that the saved recipe does not have 'block_size'.
    for config in saved_recipe:
      op_config = config.get('op_config')
      if op_config:
        weight_config = op_config.get('weight_tensor_config')
        if weight_config:
          self.assertNotIn('block_size', weight_config)

  def test_save_no_quantize_raise_error(self):
    error_message = 'No quantized model to save.'
    with self.assertRaisesWithPredicateMatch(
        RuntimeError, lambda err: error_message in str(err)
    ):
      self._quantizer._result.save(self._tmp_save_path, 'test_model')

  def test_export_model_succeeds(self):
    model_name = 'exported_model'
    self._quantizer.load_quantization_recipe(self._test_recipe_path)
    result = self._quantizer.quantize()

    exported_model_path = str(
        pathlib.Path(self._tmp_save_path) / (model_name + '.tflite')
    )
    self.assertFalse(pathlib.Path(exported_model_path).exists())
    result.export_model(exported_model_path)
    self.assertTrue(pathlib.Path(exported_model_path).exists())

  def test_compare_succeeds(self):
    self._quantizer.quantize()
    validation_result = self._quantizer.validate()
    validation_result = validation_result.get_signature_comparison_result()
    self.assertIsNotNone(validation_result)
    self.assertIn(
        'sequential/dense_1/MatMul', validation_result.intermediate_tensors
    )

  def test_validate_output_tensors_only_succeeds(self):
    self._quantizer.quantize()
    validation_result = self._quantizer.validate(
        validate_output_tensors_only=True
    )
    validation_result = validation_result.get_signature_comparison_result()
    self.assertIsNotNone(validation_result)
    self.assertEmpty(validation_result.input_tensors)
    self.assertEmpty(validation_result.constant_tensors)
    self.assertEmpty(validation_result.intermediate_tensors)
    self.assertNotEmpty(validation_result.output_tensors)
    self.assertIn('StatefulPartitionedCall:0', validation_result.output_tensors)

  def test_validate_with_quantized_model_arg_succeeds(self):
    self._quantizer.quantize()
    quantized_model = self._quantizer._result.quantized_model
    self.assertIsNotNone(quantized_model)

    new_quantizer = quantizer.Quantizer(
        self._test_model_path, previous_quantized_model=quantized_model
    )
    validation_result = new_quantizer.validate()
    validation_result = validation_result.get_signature_comparison_result()
    self.assertIsNotNone(validation_result)
    self.assertIn(
        'sequential/dense_1/MatMul', validation_result.intermediate_tensors
    )

  def test_load_custom_policies_succeeds(self):

    test_op_config = qtyping.OpQuantizationConfig(
        weight_tensor_config=_TensorQuantConfig(num_bits=4, symmetric=True),
        compute_precision=_ComputePrecision.INTEGER,
    )

    # Check if the quant config is supported by default policy.
    self._quantizer.update_quantization_recipe(
        regex='.*/Dense/.*',
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        op_config=test_op_config,
    )

    # Check if the quant config fails on dummy policy.
    dummy_policy_path = test_utils.get_path_to_datafile(
        'policies/dummy_config_policy.json'
    )
    self._quantizer.load_config_policy(dummy_policy_path)
    with self.assertRaisesRegex(
        ValueError, 'Unsupported op for .*FULLY_CONNECTED'
    ):
      self._quantizer.update_quantization_recipe(
          regex='.*/Dense/.*',
          operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
          algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
          op_config=test_op_config,
      )

    # Check if the quant config is supported by example policy in *.json file.
    default_policy_path = test_utils.get_path_to_datafile(
        'policies/example_config_policy.json'
    )
    self._quantizer.load_config_policy(default_policy_path)
    self._quantizer.update_quantization_recipe(
        regex='.*/Dense/.*',
        operation_name=qtyping.TFLOperationName.FULLY_CONNECTED,
        algorithm_key=_AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        op_config=test_op_config,
    )

    # Restore default policy to avoid side effects on other tests.
    algorithm_manager.register_config_check_policy_func(
        algorithm_manager.AlgorithmName.MIN_MAX_UNIFORM_QUANT,
        default_policy.DEFAULT_CONFIG_CHECK_POLICY,
    )

  def test_two_pass_quantization_with_conv_and_fc_succeeds(self):
    float_model_path = self._test_model_path

    drq_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/dynamic_wi8_afp32_hadamard_recipe.json'
    )
    drq_quantizer = quantizer.Quantizer(float_model_path)
    drq_quantizer.load_quantization_recipe(drq_recipe_path)
    drq_result = drq_quantizer.quantize()
    drq_model_path = str(pathlib.Path(self._tmp_save_path) / 'drq_model.tflite')
    drq_result.export_model(drq_model_path)

    srq_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'recipes/default_a8w8_recipe.json'
    )
    srq_quantizer = quantizer.Quantizer(drq_model_path)
    srq_quantizer.load_quantization_recipe(srq_recipe_path)
    representative_dataset = (
        tfl_interpreter_utils.create_random_normal_input_data(
            drq_model_path, num_samples=1
        )
    )
    calibration_result = srq_quantizer.calibrate(representative_dataset)
    srq_result = srq_quantizer.quantize(calibration_result)
    srq_model_path = str(pathlib.Path(self._tmp_save_path) / 'srq_model.tflite')
    srq_result.export_model(srq_model_path)


class QuantizerBytearrayInputs(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/conv_fc_mnist.tflite'
    )
    self._test_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/default_af32w8float_recipe.json',
    )
    with open(self._test_model_path, 'rb') as f:
      model_content = bytearray(f.read())
    with open(self._test_recipe_path, 'r') as f:
      self._test_recipe = json.load(f)
    self._quantizer = quantizer.Quantizer(model_content, self._test_recipe)

  def test_quantize_compare_succeeds(self):
    quant_result = self._quantizer.quantize()
    self.assertEqual(quant_result.recipe, self._test_recipe)
    self.assertIsNotNone(quant_result.quantized_model)
    validation_result = self._quantizer.validate()
    validation_result = validation_result.get_signature_comparison_result()
    self.assertIsNotNone(validation_result)
    self.assertIn(
        'sequential/dense_1/MatMul', validation_result.intermediate_tensors
    )

  def test_compare_succeeds(self):
    self._quantizer.quantize()
    validation_result = self._quantizer.validate()
    validation_result = validation_result.get_signature_comparison_result()
    self.assertIsNotNone(validation_result)
    self.assertIn(
        'sequential/dense_1/MatMul', validation_result.intermediate_tensors
    )


# TODO: b/364974841 - Add more tests after multiple signatures are supported
# for calibrate and quantize.
class QuantizerMultiSignatureModelTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._tmp_save_path = self.create_tempdir().full_path
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/two_signatures.tflite'
    )
    self._test_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/default_a8w8_recipe.json',
    )
    with open(self._test_recipe_path) as json_file:
      self._test_recipe = json.load(json_file)
    self._calibration_result = {
        'add_x:0': {'min': -2.0, 'max': 2.0},
        'PartitionedCall:0': {'min': -8.0, 'max': 12.0},
        'multiply_x:0': {'min': -2.0, 'max': 2.0},
        'PartitionedCall_1:0': {'min': -20.0, 'max': 20.0},
    }
    self._quantizer = quantizer.Quantizer(
        self._test_model_path, self._test_recipe_path
    )

  @parameterized.named_parameters(
      ('default_random_data', None),
      ('specific_data', _MULTI_SIGNATURE_CALIBRATION_DATASET),
  )
  def test_validate_multiple_signatures_succeeds(self, test_data):
    self._quantizer.quantize(self._calibration_result)
    validation_result = self._quantizer.validate(test_data)
    available_signatures = validation_result.available_signature_keys()
    self.assertLen(available_signatures, 2)

    add_result = validation_result.get_signature_comparison_result('add')
    self.assertEqual('mse', add_result.error_metric)
    self.assertIn('add_x:0', add_result.input_tensors)
    self.assertIn('PartitionedCall:0', add_result.output_tensors)
    self.assertIn('Add/y', add_result.constant_tensors)
    self.assertEmpty(add_result.intermediate_tensors)

    mul_result = validation_result.get_signature_comparison_result('multiply')
    self.assertEqual('mse', mul_result.error_metric)
    self.assertIn('multiply_x:0', mul_result.input_tensors)
    self.assertIn('PartitionedCall_1:0', mul_result.output_tensors)
    self.assertIn('Mul/y', mul_result.constant_tensors)
    self.assertEmpty(mul_result.intermediate_tensors)

  def test_validate_add_signature_succeeds(self):
    test_data = {'add': [{'x': np.array([2.0]).astype(np.float32)}]}
    self._quantizer.quantize(self._calibration_result)
    validation_result = self._quantizer.validate(test_data)
    available_signatures = validation_result.available_signature_keys()
    self.assertLen(available_signatures, 1)
    self.assertIn('add', available_signatures)
    add_result = validation_result.get_signature_comparison_result('add')
    self.assertEqual('mse', add_result.error_metric)
    self.assertIn('add_x:0', add_result.input_tensors)
    self.assertIn('PartitionedCall:0', add_result.output_tensors)
    self.assertIn('Add/y', add_result.constant_tensors)
    self.assertEmpty(add_result.intermediate_tensors)

  def test_validate_multiply_signature_succeeds(self):
    test_data = {'multiply': [{'x': np.array([1.0]).astype(np.float32)}]}
    self._quantizer.quantize(self._calibration_result)
    validation_result = self._quantizer.validate(test_data)
    available_signatures = validation_result.available_signature_keys()
    self.assertLen(available_signatures, 1)
    self.assertIn('multiply', available_signatures)
    mul_result = validation_result.get_signature_comparison_result('multiply')
    self.assertEqual('mse', mul_result.error_metric)
    self.assertIn('multiply_x:0', mul_result.input_tensors)
    self.assertIn('PartitionedCall_1:0', mul_result.output_tensors)
    self.assertIn('Mul/y', mul_result.constant_tensors)
    self.assertEmpty(mul_result.intermediate_tensors)

  def test_validate_quantize_after_calibration_succeeds(self):
    calib_result = self._quantizer.calibrate(
        _MULTI_SIGNATURE_CALIBRATION_DATASET
    )
    self._quantizer.quantize(calib_result)
    validation_result = self._quantizer.validate(
        _MULTI_SIGNATURE_CALIBRATION_DATASET
    )
    available_signatures = validation_result.available_signature_keys()
    self.assertLen(available_signatures, 2)

  def test_constant_buffer_shared_by_tensors_with_different_quantization_params_succeeds(
      self,
  ):
    recipe = [
        dict({
            'regex': '.*',
            'operation': 'ADD',
            'algorithm_key': 'min_max_uniform_quantize',
            'op_config': {
                'activation_tensor_config': {
                    'num_bits': 8,
                    'symmetric': False,
                    'granularity': 'TENSORWISE',
                    'dtype': 'INT',
                },
                'weight_tensor_config': {
                    'num_bits': 8,
                    'symmetric': True,
                    'granularity': 'CHANNELWISE',
                    'dtype': 'INT',
                },
                'compute_precision': 'INTEGER',
                'explicit_dequantize': False,
                'skip_checks': False,
            },
        })
    ]
    qt = quantizer.Quantizer(self._test_model_path, recipe)
    calib_result = qt.calibrate(_MULTI_SIGNATURE_CALIBRATION_DATASET)
    self.assertIsNotNone(qt.quantize(calib_result).quantized_model)

  def test_quantization_with_insufficient_calibration(self):
    # Run calibration for one signature only.
    scarce_calibration_dataset = {
        'add': [{'x': np.array([2.0], dtype=np.float32)}],
    }
    calib_result = self._quantizer.calibrate(scarce_calibration_dataset)

    # Quantize and expect an error about missing signature in calibration data.
    error_message = 'MUL(index: 0) not found in tensor_name_to_qsv'
    with self.assertRaisesWithPredicateMatch(
        ValueError, lambda err: error_message in str(err)
    ):
      self._quantizer.quantize(calib_result)


class QuantizerToyGemma2Test(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._tmp_save_path = self.create_tempdir().full_path
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'tests/models/toy_model_with_kv_cache_multi_signature.tflite',
    )

    self._toy_gemma2_calibration_dataset = {
        'signature_1': [{
            'cache_0': _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            'cache_1': _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            'positions': (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
            'tokens': (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
        }],
        'signature_2': [{
            'cache_0': _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            'cache_1': _RNG.random(size=(1, 100, 4, 4), dtype=np.float32),
            'positions': (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
            'tokens': (
                _RNG.integers(low=0, high=10, size=(1, 100)).astype(np.int32)
            ),
        }],
    }

    self._test_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/default_a8w8_recipe.json',
    )
    with open(self._test_recipe_path) as json_file:
      self._test_recipe = json.load(json_file)

    self._quantizer = quantizer.Quantizer(
        self._test_model_path, self._test_recipe_path
    )

    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=qtyping.TFLOperationName.OUTPUT,
        algorithm_key=_AlgorithmName.NO_QUANTIZE,
    )

  def test_toy_gemma2_quantization_succeeds(self):
    calib_result = self._quantizer.calibrate(
        self._toy_gemma2_calibration_dataset
    )
    self.assertIsNotNone(calib_result)
    self._quantizer.quantize(calib_result)
    self.assertIsNotNone(self._quantizer._result.quantized_model)

  def test_toy_gemma2_update_signature_defs_succeeds(self):

    self.assertTrue(
        _is_all_signature_defs_outputs_float(
            open(self._test_model_path, 'rb').read()
        )
    )
    calib_result = self._quantizer.calibrate(
        self._toy_gemma2_calibration_dataset
    )
    self.assertIsNotNone(calib_result)
    self._quantizer.quantize(calib_result)
    self.assertIsNotNone(self._quantizer._result.quantized_model)
    self.assertTrue(
        _is_all_signature_defs_outputs_float(
            self._quantizer._result.quantized_model
        )
    )


class QuantizerFullyConnectedTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._tmp_save_path = self.create_tempdir().full_path
    self._test_model_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH) / 'tests/models/single_fc.tflite',
    )

    self._test_recipe_path = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / 'recipes/default_a8w8_recipe.json',
    )
    with open(self._test_recipe_path) as json_file:
      self._test_recipe = json.load(json_file)

    self._quantizer = quantizer.Quantizer(
        self._test_model_path, self._test_recipe_path
    )

    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=qtyping.TFLOperationName.INPUT,
        algorithm_key=_AlgorithmName.NO_QUANTIZE,
    )
    self._quantizer.update_quantization_recipe(
        regex='.*',
        operation_name=qtyping.TFLOperationName.OUTPUT,
        algorithm_key=_AlgorithmName.NO_QUANTIZE,
    )

  def test_fully_connected_quantization_succeeds(self):
    calib_result = self._quantizer.calibrate(
        tfl_interpreter_utils.create_random_normal_input_data(
            self._test_model_path, num_samples=4
        )
    )
    self.assertIsNotNone(calib_result)
    self._quantizer.quantize(calib_result)
    self.assertIsNotNone(self._quantizer._result.quantized_model)

  def test_fully_connected_quantization_update_signature_defs_succeeds(self):

    model_content = open(self._test_model_path, 'rb').read()
    self.assertTrue(_is_all_signature_defs_inputs_float(model_content))
    self.assertTrue(_is_all_signature_defs_outputs_float(model_content))

    calib_result = self._quantizer.calibrate(
        tfl_interpreter_utils.create_random_normal_input_data(
            self._test_model_path, num_samples=4
        )
    )
    self.assertIsNotNone(calib_result)
    quant_result = self._quantizer.quantize(calib_result)
    self.assertIsNotNone(quant_result.quantized_model)

    self.assertTrue(
        _is_all_signature_defs_inputs_float(quant_result.quantized_model)
    )
    self.assertTrue(
        _is_all_signature_defs_outputs_float(quant_result.quantized_model)
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Model Modifier class that produce the final quantized TFlite model."""

import copy
import logging

import numpy as np

from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import transformation_instruction_generator
from ai_edge_quantizer import transformation_performer
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_litert import interpreter as tfl  # pylint: disable=g-direct-tensorflow-import


_DEQUANT_SUFFIX = "_dequant"
_QUANT_SUFFIX = "_quantized"


class ModelModifier:
  """Model Modifier class that produce the final quantized TFlite model."""

  def __init__(self, float_model: tfl_flatbuffer_utils.ModelT):
    """Constructor.

    Args:
      float_model: the original TFlite model.
    """
    self._model: tfl_flatbuffer_utils.ModelT = float_model

    self._constant_map = []
    self._transformation_instruction_generator = (
        transformation_instruction_generator.TransformationInstructionsGenerator()
    )
    self._transformation_performer = (
        transformation_performer.TransformationPerformer()
    )

  def _get_tensor_processing_order(
      self,
      tensor_names: set[str],
      flatbuffer_model: tfl_flatbuffer_utils.ModelT,
  ) -> list[str]:
    """Get the tensor processing order obtained from `buffer_to_tensors`.

    The processing order is used to ensure that last tensor in a buffer is
    processed the last. This is required for the correctness of buffer
    duplication, as the last tensor in a buffer won't be duplicated.

    Args:
      tensor_names: Names of the tensors that need to be processed.
      flatbuffer_model: TFlite model.

    Returns:
      A list of tensor names in the processing order.
    """
    buffer_to_tensors = tfl_flatbuffer_utils.buffer_to_tensors(flatbuffer_model)

    processing_order = []
    for buffer_tensors in buffer_to_tensors.values():
      for tensor in buffer_tensors:
        tensor_name = tfl_flatbuffer_utils.get_tensor_name(tensor)
        if tensor_name in tensor_names:
          processing_order.append(tensor_name)

    return processing_order

  def modify_model(
      self, params: dict[str, qtyping.TensorTransformationParams]
  ) -> bytearray:
    """Modify the model.

    Args:
      params: a dictionary with tensor name and a list of tensor transformation
        params

    Returns:
      a byte buffer that represents the serialized tflite model
    """
    # Make a copy of the model, but don't duplicate the buffer data.
    buffer_data = []
    for buffer in self._model.buffers:
      buffer_data.append(buffer.data)
      buffer.data = None
    quantized_model = copy.deepcopy(self._model)
    for bid, data in enumerate(buffer_data):
      self._model.buffers[bid].data = data
      quantized_model.buffers[bid].data = data

    instructions = self._transformation_instruction_generator.quant_params_to_transformation_insts(
        params, quantized_model
    )

    tensor_processing_order = self._get_tensor_processing_order(
        set(instructions.keys()), quantized_model
    )
    self._transformation_performer.transform_graph(
        instructions, quantized_model, tensor_processing_order
    )
    constant_buffer_size = self._process_constant_map(quantized_model)
    # we leave 256MB for the model architecture.
    serialize_fun = (
        self._serialize_large_model
        if constant_buffer_size > 2**31 - 2**28
        else self._serialize_small_model
    )
    serialized_quantized_model = serialize_fun(quantized_model)

    # Update signature defs if dequant is inserted before output.
    if self._has_transform_before_output(
        instructions, qtyping.QuantTransformation.ADD_DEQUANTIZE
    ):
      quantized_model = self._update_signature_defs(
          quantized_model, serialized_quantized_model, _DEQUANT_SUFFIX
      )
      serialized_quantized_model = serialize_fun(quantized_model)

    # Update signature defs if quant is inserted before output.
    if self._has_transform_before_output(
        instructions, qtyping.QuantTransformation.ADD_QUANTIZE
    ):
      quantized_model = self._update_signature_defs(
          quantized_model, serialized_quantized_model, _QUANT_SUFFIX
      )
      serialized_quantized_model = serialize_fun(quantized_model)

    return serialized_quantized_model

  def _update_signature_defs(
      self,
      model: tfl_flatbuffer_utils.ModelT,
      serialized_model: bytearray,
      suffix: str,
  ) -> tfl_flatbuffer_utils.ModelT:
    """Updates the signature definitions in the model.

    This function is called when a transformation (quantize or dequantize)
    is inserted before an output tensor. It updates the tensor index in the
    signature definitions to point to the newly inserted output tensor.

    Args:
      model: The TFlite ModelT object.
      serialized_model: The serialized bytearray of the TFlite model.
      suffix: The suffix to append to the tensor name.

    Returns:
      The updated TFlite ModelT object.
    """
    interpreter = tfl.Interpreter(model_content=bytes(serialized_model))

    for signature_def in model.signatureDefs:
      signature_key = signature_def.signatureKey.decode("utf-8")
      logging.info("Signature = %s", signature_key)
      subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
          interpreter, signature_key
      )
      output_details = interpreter.get_signature_runner(
          signature_key
      ).get_output_details()
      subgraph = model.subgraphs[subgraph_idx]
      graph_info = qtyping.GraphInfo(subgraph.tensors, model.buffers)

      for output in subgraph.outputs:
        tensor_name = tfl_flatbuffer_utils.get_tensor_name(
            graph_info.subgraph_tensors[output]
        )
        logging.info("\tOutput tensor = `%s`", tensor_name)

        for signature_name, tensor_details in output_details.items():
          if tensor_details["name"] + suffix == tensor_name:
            logging.info(
                "\t\tfound tensor mapping: `%s`->`%s` for signature name: `%s`",
                tensor_details["name"],
                tensor_name,
                signature_name,
            )
            for signature_item in signature_def.outputs:
              if signature_item.name.decode("utf-8") == signature_name:
                signature_item.tensorIndex = output
                logging.info(
                    "\t\t\tswapped tensor index: %s->%s",
                    tensor_details["index"],
                    output,
                )
                break
            break

    return model

  def _has_transform_before_output(
      self,
      instructions: dict[str, qtyping.TensorTransformationInsts],
      transformation: qtyping.QuantTransformation,
  ) -> bool:
    """Check if the model has transformation insert to output."""
    for tensor_name, tensor_trans_insts in instructions.items():
      for instr in tensor_trans_insts.instructions:
        if transformation == instr.transformation and instr.consumers == [-1]:
          logging.info(
              "Found %s insert to output for tensor: %s",
              transformation,
              tensor_name,
          )
          return True
    return False

  def _process_constant_map(
      self, quantized_model: tfl_flatbuffer_utils.ModelT
  ) -> int:
    """Process the constant map after all transformations are applied.

    If the resulting model is > 2GB then we would need to serialize constants
    separately, as such, we collect all the constant buffers using this
    function.

    Args:
      quantized_model: a quantized TFlite ModelT

    Returns:
      an integer representing the total size of the constant buffers
    """
    buffer_size = 0
    for buffer in quantized_model.buffers:
      if buffer.data is None:
        self._constant_map.append(buffer.data)
      elif isinstance(buffer.data, np.ndarray):
        self._constant_map.append(buffer.data.tobytes())
        buffer_size += len(buffer.data.tobytes())
      else:
        self._constant_map.append(buffer.data)
        buffer_size += len(buffer.data)
    return buffer_size

  def _pad_bytearray(self, bytearr: bytearray):
    """Pad the bytearray to 16 bytes."""
    remainder = len(bytearr) % 16
    if remainder != 0:
      padding_size = 16 - remainder
      bytearr.extend(b"\0" * padding_size)

  # TODO: b/333797307 - support > 2GB output model
  def _serialize_large_model(
      self, quantized_model: tfl_flatbuffer_utils.ModelT
  ) -> bytearray:
    """serialize models > 2GB.

    Args:
      quantized_model: a quantized TFlite ModelT

    Returns:
      a byte buffer that represents the serialized tflite model
    """
    # TODO: b/338244867 - we can have more efficient way to calculate the
    # buffer offsets.

    # remove all the constant from the model.
    for buffer in quantized_model.buffers:
      if buffer.data is not None:
        buffer.data = None
        buffer.offset = 1
        buffer.size = 1
    dummy_bytearray = bytearray(
        flatbuffer_utils.convert_object_to_bytearray(quantized_model)
    )
    # calculate the correct buffer size and offset
    self._pad_bytearray(dummy_bytearray)
    for buffer_idx, buffer in enumerate(quantized_model.buffers):
      buffer_data = self._constant_map[buffer_idx]
      if buffer_data is None:
        continue
      buffer.offset = len(dummy_bytearray)
      buffer.size = len(buffer_data)
      dummy_bytearray += buffer_data
      self._pad_bytearray(dummy_bytearray)
    del dummy_bytearray

    # build new tflite file with correct buffer offset
    model_bytearray = bytearray(
        flatbuffer_utils.convert_object_to_bytearray(quantized_model)
    )
    self._pad_bytearray(model_bytearray)
    for buffer_idx, _ in enumerate(quantized_model.buffers):
      buffer_data = self._constant_map[buffer_idx]
      if buffer_data is None:
        continue
      model_bytearray += buffer_data
      self._pad_bytearray(model_bytearray)
    return model_bytearray

  def _serialize_small_model(
      self, quantized_model: tfl_flatbuffer_utils.ModelT
  ) -> bytearray:
    """serialize models < 2GB.

    Args:
      quantized_model: a quantized TFlite ModelT

    Returns:
      a byte buffer that represents the serialized tflite model
    """
    model_bytearray = flatbuffer_utils.convert_object_to_bytearray(
        quantized_model
    )
    return model_bytearray
